import json
import logging
import tempfile
from pathlib import Path
from textwrap import dedent
from typing import Generator, get_args
from uuid import uuid4

import pytest

from ._utils.auth import _ORCA_ROOT_ACCESS_API_KEY
from .classification_model import ClassificationMetrics, ClassificationModel
from .client import OrcaClient
from .datasource import Datasource
from .embedding_model import (
    FinetunedEmbeddingModel,
    PretrainedEmbeddingModel,
    PretrainedEmbeddingModelName,
)
from .job import Status
from .memoryset import LabeledMemoryset


def test_open_pretrained_model():
    model = PretrainedEmbeddingModel.GTE_BASE
    assert model is not None
    assert isinstance(model, PretrainedEmbeddingModel)
    assert model.name == "GTE_BASE"
    assert model.embedding_dim == 768
    assert model.max_seq_length == 8192
    assert model is PretrainedEmbeddingModel.GTE_BASE


def test_open_pretrained_model_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            PretrainedEmbeddingModel.GTE_BASE.embed("I love this airline")


def test_open_pretrained_model_not_found():
    with pytest.raises(LookupError):
        PretrainedEmbeddingModel._get("INVALID_MODEL")  # type: ignore


def test_all_pretrained_models():
    models = PretrainedEmbeddingModel.all()
    assert len(models) > 1
    if len(models) != len(get_args(PretrainedEmbeddingModelName)):
        logging.warning("Please regenerate the SDK client! Some pretrained model names are not exposed yet.")
    model_names = [m.name for m in models]
    assert all(m in model_names for m in get_args(PretrainedEmbeddingModelName))


def test_embed_text():
    embedding = PretrainedEmbeddingModel.GTE_BASE.embed("I love this airline", max_seq_length=32)
    assert embedding is not None
    assert isinstance(embedding, list)
    assert len(embedding) == 768
    assert isinstance(embedding[0], float)


def test_embed_text_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            PretrainedEmbeddingModel.GTE_BASE.embed("I love this airline", max_seq_length=32)


def test_evaluate_pretrained_model(datasource: Datasource):
    metrics = PretrainedEmbeddingModel.GTE_BASE.evaluate(datasource=datasource, label_column="label")
    assert metrics is not None
    assert isinstance(metrics, ClassificationMetrics)
    assert metrics.accuracy > 0.5


def test_finetune_model_with_datasource(finetuned_embedding_model: FinetunedEmbeddingModel):
    assert finetuned_embedding_model is not None
    assert finetuned_embedding_model.name == "test_finetuned_model"
    assert finetuned_embedding_model.base_model == PretrainedEmbeddingModel.DISTILBERT
    assert finetuned_embedding_model.embedding_dim == 768
    assert finetuned_embedding_model.max_seq_length == 512
    assert finetuned_embedding_model._status == Status.COMPLETED
    # num_params should come from the base model
    assert finetuned_embedding_model.num_params == PretrainedEmbeddingModel.DISTILBERT.num_params
    assert finetuned_embedding_model.num_params == 66362880


def test_finetune_model_with_memoryset(readonly_memoryset: LabeledMemoryset):
    finetuned_embedding_model = PretrainedEmbeddingModel.DISTILBERT.finetune(
        "test_finetuned_model_from_memoryset", readonly_memoryset
    )
    assert finetuned_embedding_model is not None
    assert finetuned_embedding_model.name == "test_finetuned_model_from_memoryset"
    assert finetuned_embedding_model.base_model == PretrainedEmbeddingModel.DISTILBERT
    assert finetuned_embedding_model.embedding_dim == 768
    assert finetuned_embedding_model.max_seq_length == 512
    assert finetuned_embedding_model._status == Status.COMPLETED
    # num_params should come from the base model
    assert finetuned_embedding_model.num_params == PretrainedEmbeddingModel.DISTILBERT.num_params
    assert finetuned_embedding_model.num_params == 66362880


def test_finetune_model_already_exists_error(datasource: Datasource, finetuned_embedding_model):
    with pytest.raises(ValueError):
        PretrainedEmbeddingModel.DISTILBERT.finetune("test_finetuned_model", datasource)


def test_finetune_model_already_exists_return(datasource: Datasource, finetuned_embedding_model):
    with pytest.raises(ValueError):
        PretrainedEmbeddingModel.GTE_BASE.finetune("test_finetuned_model", datasource, if_exists="open")

    new_model = PretrainedEmbeddingModel.DISTILBERT.finetune("test_finetuned_model", datasource, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_finetuned_model"
    assert new_model.base_model == PretrainedEmbeddingModel.DISTILBERT
    assert new_model.embedding_dim == 768
    assert new_model.max_seq_length == 512
    assert new_model._status == Status.COMPLETED
    # num_params should match the base model
    assert new_model.num_params == PretrainedEmbeddingModel.DISTILBERT.num_params


def test_finetune_model_unauthenticated(unauthenticated_client, datasource: Datasource):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            PretrainedEmbeddingModel.DISTILBERT.finetune("test_finetuned_model_unauthenticated", datasource)


def test_use_finetuned_model_in_memoryset(
    datasource: Datasource, label_names: list[str], finetuned_embedding_model: FinetunedEmbeddingModel
):
    memoryset = LabeledMemoryset.create(
        "test_memoryset_finetuned_model",
        datasource=datasource,
        embedding_model=finetuned_embedding_model,
        label_names=label_names,
    )
    assert memoryset is not None
    assert memoryset.name == "test_memoryset_finetuned_model"
    assert memoryset.embedding_model == finetuned_embedding_model
    assert memoryset.length == datasource.length


def test_open_finetuned_model(finetuned_embedding_model: FinetunedEmbeddingModel):
    model = FinetunedEmbeddingModel.open(finetuned_embedding_model.name)
    assert isinstance(model, FinetunedEmbeddingModel)
    assert model.id == finetuned_embedding_model.id
    assert model.name == finetuned_embedding_model.name
    assert model.base_model == PretrainedEmbeddingModel.DISTILBERT
    assert model.embedding_dim == 768
    assert model.max_seq_length == 512
    assert model.num_params == finetuned_embedding_model.num_params
    assert model == finetuned_embedding_model


def test_embed_finetuned_model(finetuned_embedding_model: FinetunedEmbeddingModel):
    embedding = finetuned_embedding_model.embed("I love this airline")
    assert embedding is not None
    assert isinstance(embedding, list)
    assert len(embedding) == 768
    assert isinstance(embedding[0], float)


def test_all_finetuned_models(finetuned_embedding_model: FinetunedEmbeddingModel):
    models = FinetunedEmbeddingModel.all()
    assert len(models) > 0
    assert any(model.name == finetuned_embedding_model.name for model in models)


def test_all_finetuned_models_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            FinetunedEmbeddingModel.all()


def test_all_finetuned_models_unauthorized(unauthorized_client, finetuned_embedding_model: FinetunedEmbeddingModel):
    with unauthorized_client.use():
        assert finetuned_embedding_model not in FinetunedEmbeddingModel.all()


def test_drop_finetuned_model(datasource: Datasource):
    PretrainedEmbeddingModel.DISTILBERT.finetune("finetuned_model_to_delete", datasource)
    assert FinetunedEmbeddingModel.open("finetuned_model_to_delete")
    FinetunedEmbeddingModel.drop("finetuned_model_to_delete")
    with pytest.raises(LookupError):
        FinetunedEmbeddingModel.open("finetuned_model_to_delete")


def test_drop_finetuned_model_with_memoryset_cascade(datasource: Datasource, label_names: list[str]):
    finetuned_embedding_model = PretrainedEmbeddingModel.DISTILBERT.finetune(
        "finetuned_model_cascade_delete", datasource
    )
    memoryset = LabeledMemoryset.create(
        "test_memoryset_for_finetuned_model_cascade",
        datasource=datasource,
        embedding_model=finetuned_embedding_model,
        label_names=label_names,
    )

    # Verify memoryset exists and uses the finetuned model
    assert LabeledMemoryset.open(memoryset.name) is not None
    assert memoryset.embedding_model == finetuned_embedding_model

    # Without cascade, deletion should fail
    with pytest.raises(RuntimeError):
        FinetunedEmbeddingModel.drop(finetuned_embedding_model.id, cascade=False)

    # Model and memoryset should still exist
    assert FinetunedEmbeddingModel.exists(finetuned_embedding_model.name)
    assert LabeledMemoryset.exists(memoryset.name)

    # With cascade, deletion should succeed
    FinetunedEmbeddingModel.drop(finetuned_embedding_model.id, cascade=True)

    # Both model and memoryset should be deleted
    assert not FinetunedEmbeddingModel.exists(finetuned_embedding_model.name)
    assert not LabeledMemoryset.exists(memoryset.name)


def test_drop_finetuned_model_unauthenticated(unauthenticated_client, datasource: Datasource):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            PretrainedEmbeddingModel.DISTILBERT.finetune("finetuned_model_to_delete", datasource)


def test_drop_finetuned_model_not_found():
    with pytest.raises(LookupError):
        FinetunedEmbeddingModel.drop(str(uuid4()))
    # ignores error if specified
    FinetunedEmbeddingModel.drop(str(uuid4()), if_not_exists="ignore")


def test_drop_finetuned_model_unauthorized(unauthorized_client, finetuned_embedding_model: FinetunedEmbeddingModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            FinetunedEmbeddingModel.drop(finetuned_embedding_model.id)


def test_supports_instructions():
    model = PretrainedEmbeddingModel.GTE_BASE
    assert not model.supports_instructions

    instruction_model = PretrainedEmbeddingModel.BGE_BASE
    assert instruction_model.supports_instructions


def test_use_explicit_instruction_prompt():
    model = PretrainedEmbeddingModel.BGE_BASE
    assert model.supports_instructions
    input = "Hello world"
    assert model.embed(input, prompt="Represent this sentence for sentiment retrieval:") != model.embed(input)


def test_pretrained_embedding_model_repr():
    assert repr(PretrainedEmbeddingModel.GTE_BASE) == (
        "PretrainedEmbeddingModel({ name: GTE_BASE, embedding_dim: 768, max_seq_length: 8192, num_params: 137M })"
    )


def test_finetuned_embedding_model_repr(finetuned_embedding_model: FinetunedEmbeddingModel):
    assert repr(finetuned_embedding_model) == dedent(
        """\
        FinetunedEmbeddingModel({
            name: test_finetuned_model,
            embedding_dim: 768,
            max_seq_length: 512,
            base_model: PretrainedEmbeddingModel.DISTILBERT
        })"""
    )


@pytest.fixture(scope="module")
def root_client(org_id) -> Generator[OrcaClient, None, None]:
    """Client with root API key access for testing privileged endpoints."""
    client = OrcaClient(api_key=_ORCA_ROOT_ACCESS_API_KEY, headers={"Org-Id": org_id})
    yield client


@pytest.fixture(scope="module")
def mock_model_dir() -> Generator[Path, None, None]:
    """
    Create a minimal mock HuggingFace model directory for upload testing.

    The config.json contains the minimum fields required for EmbeddingModel.load_config
    to extract embedding dimension and max sequence length.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        model_dir = Path(temp_dir)

        # Create a minimal HuggingFace-style config.json
        config = {
            "model_type": "bert",
            "architectures": ["BertModel"],
            "hidden_size": 256,  # This becomes embedding_dim
            "max_position_embeddings": 128,  # This becomes max_seq_length
            "vocab_size": 30522,
            "num_hidden_layers": 2,
            "num_attention_heads": 4,
            "intermediate_size": 512,
        }
        config_path = model_dir / "config.json"
        config_path.write_text(json.dumps(config, indent=2))

        yield model_dir


def test_upload_embedding_model(root_client: OrcaClient, mock_model_dir: Path):
    """Test uploading a pretrained embedding model from a local directory."""
    model_name = f"test_uploaded_model_{uuid4().hex[:8]}"
    description = "Test uploaded embedding model"

    with root_client.use():
        try:
            # Upload the model
            model = FinetunedEmbeddingModel._upload(
                model_name,
                mock_model_dir,
                description=description,
            )

            # Verify the model was created with correct metadata
            assert model.name == model_name
            assert model.embedding_dim == 256  # from config hidden_size
            assert model.max_seq_length == 128  # from config max_position_embeddings
            assert model.base_model is None  # Uploaded models have no base model
            assert model.description == description
            # num_params won't be abe able to be extracted from the model
            assert model.num_params is None or model.num_params == 0

            # Verify we can open it by name
            opened_model = FinetunedEmbeddingModel.open(model_name)
            assert opened_model.id == model.id
            assert opened_model.base_model is None
            assert opened_model.description == description

            # Verify it appears in the list
            all_models = FinetunedEmbeddingModel.all()
            assert any(m.id == model.id for m in all_models)

        finally:
            # Clean up
            FinetunedEmbeddingModel.drop(model_name, if_not_exists="ignore")


def test_upload_embedding_model_already_exists_error(root_client: OrcaClient, mock_model_dir: Path):
    """Test that uploading with an existing name raises an error by default."""
    model_name = f"test_uploaded_model_exists_{uuid4().hex[:8]}"

    with root_client.use():
        try:
            # First upload
            FinetunedEmbeddingModel._upload(model_name, mock_model_dir, description="First description")

            # Second upload should fail
            with pytest.raises(ValueError, match="already exists"):
                FinetunedEmbeddingModel._upload(model_name, mock_model_dir, description="Second description")

        finally:
            FinetunedEmbeddingModel.drop(model_name, if_not_exists="ignore")


def test_upload_embedding_model_already_exists_open(root_client: OrcaClient, mock_model_dir: Path):
    """Test that uploading with if_exists='open' returns the existing model."""
    model_name = f"test_uploaded_model_open_{uuid4().hex[:8]}"
    description = "Original description"

    with root_client.use():
        try:
            # First upload
            original = FinetunedEmbeddingModel._upload(model_name, mock_model_dir, description=description)

            # Second upload with if_exists="open" should return the existing model
            existing = FinetunedEmbeddingModel._upload(model_name, mock_model_dir, if_exists="open")
            assert existing.id == original.id
            assert existing.description == description

        finally:
            FinetunedEmbeddingModel.drop(model_name, if_not_exists="ignore")


def test_upload_embedding_model_without_description(root_client: OrcaClient, mock_model_dir: Path):
    """Test that uploading without a description works correctly."""
    model_name = f"test_uploaded_model_no_desc_{uuid4().hex[:8]}"

    with root_client.use():
        try:
            # Upload the model without description
            model = FinetunedEmbeddingModel._upload(model_name, mock_model_dir)

            # Verify description is None
            assert model.description is None

            # Verify we can open it and description is still None
            opened_model = FinetunedEmbeddingModel.open(model_name)
            assert opened_model.description is None
            # num_params should be consistent
            assert opened_model.num_params == model.num_params

        finally:
            FinetunedEmbeddingModel.drop(model_name, if_not_exists="ignore")


def test_upload_embedding_model_invalid_path(root_client: OrcaClient):
    """Test that uploading from a non-directory path raises an error."""
    with root_client.use():
        with pytest.raises(ValueError, match="must be a directory"):
            FinetunedEmbeddingModel._upload("test_model", "/nonexistent/path")


def test_upload_embedding_model_unauthorized(mock_model_dir: Path):
    """Test that uploading without root access raises an error."""
    # The regular api_key fixture doesn't have root access
    with pytest.raises(PermissionError):
        FinetunedEmbeddingModel._upload("test_model_unauthorized", mock_model_dir)


def test_upload_gte_small_and_create_classification_model(root_client: OrcaClient):
    """
    Integration test: upload GTE-Small, create memoryset and classification model, verify predictions.

    This test mirrors the functionality in research/debug/upload_embedding_model.ipynb:
    1. Download and upload a real HuggingFace embedding model (GTE-Small)
    2. Verify embeddings work
    3. Create a labeled memoryset with the uploaded model
    4. Create a classification model
    5. Make predictions and verify they work correctly
    """
    hf_hub = pytest.importorskip("huggingface_hub")

    unique_suffix = uuid4().hex[:8]
    model_name = f"test_uploaded_gte_small_{unique_suffix}"
    memoryset_name = f"test_uploaded_gte_memoryset_{unique_suffix}"
    classifier_name = f"test_uploaded_gte_classifier_{unique_suffix}"

    with tempfile.TemporaryDirectory() as temp_dir:
        # Download GTE-Small from HuggingFace (only safetensors, not ONNX)
        model_dir = Path(temp_dir)
        hf_hub.snapshot_download(
            repo_id="thenlper/gte-small",
            local_dir=model_dir,
            allow_patterns=[
                "*.safetensors",
                "*.json",
                "*.txt",
                "*.py",
                "*.model",
            ],  # Include safetensors and config/tokenizer files
            ignore_patterns=["*.onnx"],  # Exclude ONNX files
        )

        with root_client.use():
            try:
                # Upload the model
                model = FinetunedEmbeddingModel._upload(
                    model_name,
                    model_dir,
                    description="GTE-Small model for integration testing",
                )

                # Verify model properties
                assert model.name == model_name
                assert model.embedding_dim == 384  # GTE-Small has 384 dimensions
                assert model.max_seq_length == 512
                assert model.base_model is None  # Uploaded models have no base model
                assert model.description == "GTE-Small model for integration testing"
                # NOTE: currently num_params cannot be reliably determined for uploaded models
                assert model.num_params == 33360512

                # Test embedding functionality
                embedding = model.embed("This is a test sentence for embedding.")
                assert isinstance(embedding, list)
                assert len(embedding) == 384
                assert isinstance(embedding[0], float)

                # Create labeled memoryset with sample sentiment data
                sample_data = [
                    {"value": "I love this product, it's amazing!", "label": 0},
                    {"value": "Great experience, highly recommend", "label": 0},
                    {"value": "Best purchase I've ever made", "label": 0},
                    {"value": "Wonderful service and quality", "label": 0},
                    {"value": "Exceeded all my expectations", "label": 0},
                    {"value": "This is terrible, waste of money", "label": 1},
                    {"value": "Horrible experience, never again", "label": 1},
                    {"value": "Worst product I've ever bought", "label": 1},
                    {"value": "Very disappointing quality", "label": 1},
                    {"value": "Complete garbage, don't buy", "label": 1},
                ]

                memoryset = LabeledMemoryset.create(
                    name=memoryset_name,
                    embedding_model=model,
                    label_names=["positive", "negative"],
                )
                memoryset.insert(sample_data)

                assert memoryset.length == len(sample_data)
                assert memoryset.embedding_model == model
                assert memoryset.label_names == ["positive", "negative"]

                # Test search on memoryset
                search_results = memoryset.search("This product is fantastic!", count=3)
                assert len(search_results) == 3
                assert search_results[0].label == 0  # Top result should be positive

                # Create classification model
                classifier = ClassificationModel.create(
                    name=classifier_name,
                    memoryset=memoryset,
                )
                assert classifier.name == classifier_name

                # Test predictions
                positive_prediction = classifier.predict("This is absolutely wonderful!")
                assert positive_prediction.label_name == "positive"

                negative_prediction = classifier.predict("I hate this, it's the worst")
                assert negative_prediction.label_name == "negative"

                # Verify memory lookups are populated
                assert len(classifier.last_prediction.memory_lookups) > 0

            finally:
                # Cleanup - cascade deletes handle dependencies
                FinetunedEmbeddingModel.drop(model_name, if_not_exists="ignore", cascade=True)



import pytest

from .classification_model import ClassificationModel
from .memoryset import LabeledMemoryLookup
from .telemetry import (
    ClassificationPrediction,
    FeedbackCategory,
    LabeledMemorySuggestion,
)


def test_get_prediction(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    prediction_with_telemetry = ClassificationPrediction.get(predictions[0].prediction_id)
    assert prediction_with_telemetry is not None
    assert prediction_with_telemetry.label == 0
    assert prediction_with_telemetry.input_value == "Do you love soup?"


def test_get_predictions(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    prediction_with_telemetry = ClassificationPrediction.get(
        [predictions[0].prediction_id, predictions[1].prediction_id]
    )
    assert len(prediction_with_telemetry) == 2
    assert prediction_with_telemetry[0].label == 0
    assert prediction_with_telemetry[0].input_value == "Do you love soup?"
    assert prediction_with_telemetry[1].label == 1


def test_get_predictions_with_expected_label_match(classification_model: ClassificationModel):
    classification_model.predict(
        ["Do you love soup?", "Are cats cute?"], expected_labels=[0, 0], tags={"expected_label_match"}
    )
    classification_model.predict("no expectations", tags={"expected_label_match"})
    assert len(classification_model.predictions(tag="expected_label_match")) == 3
    assert len(classification_model.predictions(expected_label_match=True, tag="expected_label_match")) == 1
    assert len(classification_model.predictions(expected_label_match=False, tag="expected_label_match")) == 1


def test_get_prediction_memory_lookups(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    assert isinstance(prediction.memory_lookups, list)
    assert len(prediction.memory_lookups) > 0
    assert all(isinstance(lookup, LabeledMemoryLookup) for lookup in prediction.memory_lookups)


def test_record_feedback(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    assert "correct" not in prediction.feedback
    prediction.record_feedback(category="correct", value=prediction.label == 0)
    assert prediction.feedback["correct"] is True


def test_record_feedback_with_invalid_value(classification_model: ClassificationModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.predict("Do you love soup?").record_feedback(category="correct", value="not a bool")  # type: ignore


def test_record_feedback_with_inconsistent_value_for_category(classification_model: ClassificationModel):
    classification_model.predict("Do you love soup?").record_feedback(category="correct", value=True)
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.predict("Do you love soup?").record_feedback(category="correct", value=-1.0)


def test_delete_feedback(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    prediction.record_feedback(category="test_delete", value=True)
    assert "test_delete" in prediction.feedback
    prediction.delete_feedback("test_delete")
    assert "test_delete" not in prediction.feedback


def test_list_feedback_categories(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    prediction.record_feedback(category="correct", value=True)
    prediction.record_feedback(category="confidence", value=0.8)
    categories = FeedbackCategory.all()
    assert len(categories) >= 2
    assert any(c.name == "correct" and c.value_type == bool for c in categories)
    assert any(c.name == "confidence" and c.value_type == float for c in categories)


def test_drop_feedback_category(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    prediction.record_feedback(category="test_category", value=True)
    assert any(c.name == "test_category" for c in FeedbackCategory.all())
    FeedbackCategory.drop("test_category")
    assert not any(c.name == "test_category" for c in FeedbackCategory.all())
    prediction.refresh()
    assert "test_category" not in prediction.feedback


def test_update_prediction(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    assert prediction.expected_label is None
    assert prediction.tags == set()
    # update expected label
    prediction.update(expected_label=1)
    assert prediction.expected_label == 1

    # update tags
    prediction.update(tags={"test_tag1", "test_tag2"})
    assert prediction.tags == {"test_tag1", "test_tag2"}

    # update both
    prediction.update(expected_label=0, tags={"new_tag"})
    assert prediction.expected_label == 0
    assert prediction.tags == {"new_tag"}

    # remove expected label
    prediction.update(expected_label=None)
    assert prediction.expected_label is None

    # remove tags
    prediction.update(tags=None)
    assert prediction.tags == set()


def test_feedback_category_repr():
    category = FeedbackCategory(
        {
            "id": "cat-id",
            "org_id": "org-id",
            "name": "accuracy",
            "type": "BINARY",
            "created_at": "2025-01-01T00:00:00",
            "updated_at": "2025-01-01T00:00:00",
        }
    )
    assert repr(category) == "FeedbackCategory({ name: accuracy, value_type: <class 'bool'> })"


def test_labeled_memory_suggestion_repr():
    suggestion = LabeledMemorySuggestion(value="i love soup", label=0, label_name="soup")
    assert repr(suggestion) == "LabeledMemorySuggestion({ label: <soup: 0>, value: 'i love soup' })"



from __future__ import annotations

import tempfile
import zipfile
from datetime import datetime
from io import BytesIO
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, Union, cast

from httpx._types import FileTypes  # type: ignore
from tqdm.auto import tqdm

from ._utils.common import CreateMode, DropMode, logger
from ._utils.torch_parsing import list_from_torch
from ._utils.tqdm_file_reader import TqdmFileReader
from .client import DatasourceMetadata, OrcaClient

if TYPE_CHECKING:
    # These are peer dependencies that are used for types only
    from datasets import Dataset as HFDataset  # type: ignore
    from datasets import DatasetDict as HFDatasetDict  # type: ignore
    from pandas import DataFrame as PandasDataFrame  # type: ignore
    from pyarrow import Table as PyArrowTable  # type: ignore
    from torch.utils.data import DataLoader as TorchDataLoader  # type: ignore
    from torch.utils.data import Dataset as TorchDataset  # type: ignore


def _upload_files_to_datasource(
    name: str,
    file_paths: list[Path],
    description: str | None = None,
) -> DatasourceMetadata:
    """
    Helper function to upload files to create a datasource using manual HTTP requests.

    This bypasses the generated client because it doesn't handle file uploads properly.

    Params:
        name: Name for the datasource
        file_paths: List of file paths to upload
        description: Optional description for the datasource

    Returns:
        Metadata for the created datasource
    """
    files: list[tuple[Literal["files"], FileTypes]] = []

    # Calculate total size for all files
    total_size = sum(file_path.stat().st_size for file_path in file_paths)

    with tqdm(total=total_size, unit="B", unit_scale=True, desc="upload") as pbar:
        for file_path in file_paths:
            buffered_reader = open(file_path, "rb")
            tqdm_reader = TqdmFileReader(buffered_reader, pbar)
            files.append(("files", (file_path.name, cast(bytes, tqdm_reader))))

        # Use manual HTTP request for file uploads
        client = OrcaClient._resolve_client()
        metadata = client.POST(
            "/datasource/upload",
            files=files,
            data={"name": name, "description": description},
        )

    return metadata


def _handle_existing_datasource(name: str, if_exists: CreateMode) -> Union["Datasource", None]:
    """
    Helper function to handle the common pattern of checking if a datasource exists
    and taking action based on the if_exists parameter.

    Params:
        name: Name of the datasource to check
        if_exists: What to do if a datasource with the same name already exists

    Returns:
        Datasource instance if opening existing, None if should proceed with creation

    Raises:
        ValueError: If the datasource already exists and if_exists is "error"
    """
    if Datasource.exists(name):
        if if_exists == "error":
            raise ValueError(f"Dataset with name {name} already exists")
        elif if_exists == "open":
            return Datasource.open(name)
    return None


class Datasource:
    """
    A Handle to a datasource in the OrcaCloud

    A Datasource is a collection of data saved to the OrcaCloud that can be used to create a [`Memoryset`][orca_sdk.LabeledMemoryset].
    It can be created from a Hugging Face Dataset, a PyTorch DataLoader or Dataset, a list of dictionaries, a dictionary of columns, a pandas DataFrame, a pyarrow Table, or a local file.

    Attributes:
        id: Unique identifier for the datasource
        name: Unique name of the datasource
        description: Optional description of the datasource
        length: Number of rows in the datasource
        created_at: When the datasource was created
        columns: Dictionary of column names and types
    """

    id: str
    name: str
    description: str | None
    length: int
    created_at: datetime
    updated_at: datetime
    columns: dict[str, str]

    def __init__(self, metadata: DatasourceMetadata):
        # for internal use only, do not document
        self.id = metadata["id"]
        self.name = metadata["name"]
        self.length = metadata["length"]
        self.created_at = datetime.fromisoformat(metadata["created_at"])
        self.updated_at = datetime.fromisoformat(metadata["updated_at"])
        self.description = metadata["description"]
        self.columns = {
            column["name"]: (
                f"enum({', '.join(f'{option!r}' for option in column['enum_options'] or []) if 'enum_options' in column else ''})"
                if column["type"] == "ENUM"
                else "str" if column["type"] == "STRING" else column["type"].lower()
            )
            for column in metadata["columns"]
        }

    def __eq__(self, other) -> bool:
        return isinstance(other, Datasource) and self.id == other.id

    def __repr__(self) -> str:
        return (
            "Datasource({\n"
            + f"    name: '{self.name}',\n"
            + f"    length: {self.length},\n"
            + "    columns: ({\n        "
            + "\n        ".join([f"{k}: {v}" for k, v in self.columns.items()])
            + "\n    })\n"
            + "})"
        )

    @classmethod
    def from_hf_dataset(
        cls, name: str, dataset: HFDataset, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a Hugging Face Dataset

        Params:
            name: Required name for the new datasource (must be unique)
            dataset: The Hugging Face Dataset to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        with tempfile.TemporaryDirectory() as tmp_dir:
            dataset.save_to_disk(tmp_dir)

            # Get all file paths in the directory
            file_paths = list(Path(tmp_dir).iterdir())

            # Use the helper function to upload files
            metadata = _upload_files_to_datasource(name, file_paths, description)
            return cls(metadata=metadata)

    @classmethod
    def from_hf_dataset_dict(
        cls,
        name: str,
        dataset_dict: HFDatasetDict,
        if_exists: CreateMode = "error",
        description: dict[str, str | None] | str | None = None,
    ) -> dict[str, Datasource]:
        """
        Create datasources from a Hugging Face DatasetDict

        Params:
            name: Name prefix for the new datasources, will be suffixed with the dataset name
            dataset_dict: The Hugging Face DatasetDict to create the datasources from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasources, can be a string or a dictionary of dataset names to descriptions

        Returns:
            A dictionary of datasource handles, keyed by the dataset name

        Raises:
            ValueError: If a datasource already exists and if_exists is `"error"`
        """
        if description is None or isinstance(description, str):
            description = {str(dataset_name): description for dataset_name in dataset_dict.keys()}
        return {
            str(dataset_name): cls.from_hf_dataset(
                f"{name}_{dataset_name}", dataset, if_exists=if_exists, description=description[str(dataset_name)]
            )
            for dataset_name, dataset in dataset_dict.items()
        }

    @classmethod
    def from_pytorch(
        cls,
        name: str,
        torch_data: TorchDataLoader | TorchDataset,
        column_names: list[str] | None = None,
        if_exists: CreateMode = "error",
        description: str | None = None,
    ) -> Datasource:
        """
        Create a new datasource from a PyTorch DataLoader or Dataset

        Params:
            name: Required name for the new datasource (must be unique)
            torch_data: The PyTorch DataLoader or Dataset to create the datasource from
            column_names: If the provided dataset or data loader returns unnamed tuples, this
                argument must be provided to specify the names of the columns.
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        data_list = list_from_torch(torch_data, column_names=column_names)
        return cls.from_list(name, data_list, if_exists=if_exists, description=description)

    @classmethod
    def from_list(
        cls, name: str, data: list[dict], if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a list of dictionaries

        Params:
            name: Required name for the new datasource (must be unique)
            data: The list of dictionaries to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`

        Examples:
            >>> Datasource.from_list("my_datasource", [{"text": "Hello, world!", "label": 1}, {"text": "Goodbye", "label": 0}])
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        client = OrcaClient._resolve_client()
        metadata = client.POST(
            "/datasource",
            json={"name": name, "description": description, "content": data},
        )
        return cls(metadata=metadata)

    @classmethod
    def from_dict(
        cls, name: str, data: dict, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a dictionary of columns

        Params:
            name: Required name for the new datasource (must be unique)
            data: The dictionary of columns to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`

        Examples:
            >>> Datasource.from_dict("my_datasource", {"text": ["Hello, world!", "Goodbye"], "label": [1, 0]})
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        client = OrcaClient._resolve_client()
        metadata = client.POST(
            "/datasource",
            json={"name": name, "description": description, "content": data},
        )
        return cls(metadata=metadata)

    @classmethod
    def from_pandas(
        cls, name: str, dataframe: PandasDataFrame, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a pandas DataFrame

        Params:
            name: Required name for the new datasource (must be unique)
            dataframe: The pandas DataFrame to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Notes:
            Data type precision may be lost during upload unless the [`datasets`][datasets] library is installed.

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
            ImportError: If the upload dependency group is not installed
        """
        try:
            from datasets import Dataset  # type: ignore

            return cls.from_hf_dataset(
                name, Dataset.from_pandas(dataframe), if_exists=if_exists, description=description
            )
        except ImportError:
            return cls.from_dict(name, dataframe.to_dict(orient="list"), if_exists=if_exists, description=description)

    @classmethod
    def from_arrow(
        cls, name: str, pyarrow_table: PyArrowTable, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a pyarrow Table

        Params:
            name: Required name for the new datasource (must be unique)
            pyarrow_table: The pyarrow Table to create the datasource from
            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        # peer dependency that is guaranteed to exist if the user provided a pyarrow table
        from pyarrow import parquet  # type: ignore

        # Write to bytes buffer
        buffer = BytesIO()
        parquet.write_table(pyarrow_table, buffer)
        parquet_bytes = buffer.getvalue()

        client = OrcaClient._resolve_client()
        metadata = client.POST(
            "/datasource/upload",
            files=[("files", ("data.parquet", parquet_bytes))],
            data={"name": name, "description": description},
        )

        return cls(metadata=metadata)

    @classmethod
    def from_disk(
        cls, name: str, file_path: str | PathLike, if_exists: CreateMode = "error", description: str | None = None
    ) -> Datasource:
        """
        Create a new datasource from a local file

        Params:
            name: Required name for the new datasource (must be unique)
            file_path: Path to the file on disk to create the datasource from. The file type will
                be inferred from the file extension. The following file types are supported:

                - .pkl: [`Pickle`][pickle] files containing lists of dictionaries or dictionaries of columns
                - .json/.jsonl: [`JSON`][json] and [`JSON`] Lines files
                - .csv: [`CSV`][csv] files
                - .parquet: [`Parquet`][pyarrow.parquet.ParquetFile] files
                - dataset directory: Directory containing a saved HuggingFace [`Dataset`][datasets.Dataset]

            if_exists: What to do if a datasource with the same name already exists, defaults to
                `"error"`. Other option is `"open"` to open the existing datasource.
            description: Optional description for the datasource

        Returns:
            A handle to the new datasource in the OrcaCloud

        Raises:
            ValueError: If the datasource already exists and if_exists is `"error"`
            ImportError: If the path is a directory and [`datasets`][datasets] is not installed
        """
        # Check if datasource already exists and handle accordingly
        existing = _handle_existing_datasource(name, if_exists)
        if existing is not None:
            return existing

        file_path = Path(file_path)

        # For dataset directories, use the upload endpoint with multiple files
        if file_path.is_dir():
            try:
                from datasets import Dataset  # type: ignore
            except ImportError as e:
                raise ImportError(
                    "The path is a directory, we only support uploading directories that contain saved HuggingFace datasets but datasets is not installed."
                ) from e

            return cls.from_hf_dataset(
                name, Dataset.load_from_disk(file_path), if_exists=if_exists, description=description
            )

        # For single files, use the helper function to upload files
        metadata = _upload_files_to_datasource(name, [file_path], description)

        return cls(metadata=metadata)

    @classmethod
    def open(cls, name_or_id: str) -> Datasource:
        """
        Get a handle to a datasource by name or id in the OrcaCloud

        Params:
            name_or_id: The name or unique identifier of the datasource to get

        Returns:
            A handle to the existing datasource in the OrcaCloud

        Raises:
            LookupError: If the datasource does not exist
        """
        client = OrcaClient._resolve_client()
        return cls(client.GET("/datasource/{name_or_id}", params={"name_or_id": name_or_id}))

    @classmethod
    def exists(cls, name_or_id: str) -> bool:
        """
        Check if a datasource exists in the OrcaCloud

        Params:
            name_or_id: The name or id of the datasource to check

        Returns:
            `True` if the datasource exists, `False` otherwise
        """
        try:
            cls.open(name_or_id)
            return True
        except LookupError:
            return False

    @classmethod
    def all(cls) -> list[Datasource]:
        """
        List all datasource handles in the OrcaCloud

        Returns:
            A list of all datasource handles in the OrcaCloud
        """
        client = OrcaClient._resolve_client()
        return [cls(metadata) for metadata in client.GET("/datasource")]

    @classmethod
    def drop(cls, name_or_id: str, if_not_exists: DropMode = "error") -> None:
        """
        Delete a datasource from the OrcaCloud

        Params:
            name_or_id: The name or id of the datasource to delete
            if_not_exists: What to do if the datasource does not exist, defaults to
                `"error"`. Other options are `"ignore"` to do nothing.

        Raises:
            LookupError: If the datasource does not exist and if_not_exists is `"error"`
        """
        try:
            client = OrcaClient._resolve_client()
            client.DELETE("/datasource/{name_or_id}", params={"name_or_id": name_or_id})
            logger.info(f"Deleted datasource {name_or_id}")
        except LookupError:
            if if_not_exists == "error":
                raise

    def __len__(self) -> int:
        return self.length

    def query(
        self,
        offset: int = 0,
        limit: int = 100,
        shuffle: bool = False,
        shuffle_seed: int | None = None,
        filters: list[tuple[str, Literal["==", "!=", ">", ">=", "<", "<=", "in", "not in", "like"], Any]] = [],
    ) -> list[dict[str, Any]]:
        """
        Query the datasource for rows with pagination and filtering support.

        Params:
            offset: Number of rows to skip
            limit: Maximum number of rows to return
            shuffle: Whether to shuffle the dataset before pagination
            shuffle_seed: Seed for shuffling (for reproducible results)
            filters: List of filter tuples. Each tuple contains:
                - field (str): Column name to filter on
                - op (str): Operator ("==", "!=", ">", ">=", "<", "<=", "in", "not in", "like")
                - value: Value to compare against

        Returns:
            List of rows from the datasource

        Examples:
            >>> datasource.query(filters=[("age", ">", 25)])
            >>> datasource.query(filters=[("city", "in", ["NYC", "LA"])])
            >>> datasource.query(filters=[("name", "like", "John")])
        """

        client = OrcaClient._resolve_client()
        response = client.POST(
            "/datasource/{name_or_id}/rows",
            params={"name_or_id": self.id},
            json={
                "limit": limit,
                "offset": offset,
                "shuffle": shuffle,
                "shuffle_seed": shuffle_seed,
                "filters": [{"field": field, "op": op, "value": value} for field, op, value in filters],
            },
        )
        return response

    def download(self, output_dir: str | PathLike, file_type: Literal["hf_dataset", "json", "csv"]) -> None:
        """
        Download the datasource to a specified path in the specified format type

        Params:
            output_dir: The local directory where the downloaded file will be saved.
            file_type: The type of file to download.

        Returns:
            None
        """
        if not file_type:
            raise ValueError("file_type is required")
        extension = "zip" if file_type == "hf_dataset" else file_type
        output_path = Path(output_dir) / f"{self.name}.{extension}"
        with open(output_path, "wb") as download_file:
            client = OrcaClient._resolve_client()
            with client.stream("GET", f"/datasource/{self.id}/download", params={"file_type": file_type}) as response:
                total_chunks = int(response.headers["X-Total-Chunks"]) if "X-Total-Chunks" in response.headers else None
                with tqdm(desc="download", total=total_chunks, disable=total_chunks is None) as progress:
                    for chunk in response.iter_bytes():
                        download_file.write(chunk)
                        progress.update(1)

        # extract the zip file
        if extension == "zip":
            extract_dir = Path(output_dir) / self.name
            with zipfile.ZipFile(output_path, "r") as zip_ref:
                zip_ref.extractall(extract_dir)
            output_path.unlink()  # Remove the zip file after extraction
            logger.info(f"Downloaded {extract_dir}")
        else:
            logger.info(f"Downloaded {output_path}")

    def to_list(self) -> list[dict]:
        """
        Convert the datasource to a list of dictionaries.

        Returns:
            A list of dictionaries representation of the datasource.
        """
        client = OrcaClient._resolve_client()
        return client.GET("/datasource/{name_or_id}/download", params={"name_or_id": self.id, "file_type": "json"})



from textwrap import dedent
from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import ScoredMemoryset
from .regression_model import RegressionMetrics, RegressionModel
from .telemetry import RegressionPrediction


def test_create_model(regression_model: RegressionModel, scored_memoryset: ScoredMemoryset):
    assert regression_model is not None
    assert regression_model.name == "test_regression_model"
    assert regression_model.memoryset == scored_memoryset
    assert regression_model.memory_lookup_count == 3


def test_create_model_already_exists_error(scored_memoryset, regression_model: RegressionModel):
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset)
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset, if_exists="error")


def test_create_model_already_exists_return(scored_memoryset, regression_model: RegressionModel):
    with pytest.raises(ValueError):
        RegressionModel.create("test_regression_model", scored_memoryset, if_exists="open", memory_lookup_count=37)

    new_model = RegressionModel.create("test_regression_model", scored_memoryset, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_regression_model"
    assert new_model.memoryset == scored_memoryset
    assert new_model.memory_lookup_count == 3


def test_create_model_unauthenticated(unauthenticated_client, scored_memoryset: ScoredMemoryset):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            RegressionModel.create("test_regression_model", scored_memoryset)


def test_get_model(regression_model: RegressionModel):
    fetched_model = RegressionModel.open(regression_model.name)
    assert fetched_model is not None
    assert fetched_model.id == regression_model.id
    assert fetched_model.name == regression_model.name
    assert fetched_model.memory_lookup_count == 3
    assert fetched_model == regression_model


def test_get_model_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            RegressionModel.open("test_regression_model")


def test_get_model_invalid_input():
    with pytest.raises(ValueError, match="Invalid input"):
        RegressionModel.open("not valid id")


def test_get_model_not_found():
    with pytest.raises(LookupError):
        RegressionModel.open(str(uuid4()))


def test_get_model_unauthorized(unauthorized_client, regression_model: RegressionModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            RegressionModel.open(regression_model.name)


def test_list_models(regression_model: RegressionModel):
    models = RegressionModel.all()
    assert len(models) > 0
    assert any(model.name == regression_model.name for model in models)


def test_list_models_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            RegressionModel.all()


def test_list_models_unauthorized(unauthorized_client, regression_model: RegressionModel):
    with unauthorized_client.use():
        assert RegressionModel.all() == []


def test_memoryset_regression_models_property(regression_model: RegressionModel, scored_memoryset: ScoredMemoryset):
    models = scored_memoryset.regression_models
    assert len(models) > 0
    assert any(model.id == regression_model.id for model in models)


def test_update_model_attributes(regression_model: RegressionModel):
    regression_model.description = "New description"
    assert regression_model.description == "New description"

    regression_model.set(description=None)
    assert regression_model.description is None

    regression_model.set(locked=True)
    assert regression_model.locked is True

    regression_model.set(locked=False)
    assert regression_model.locked is False

    regression_model.lock()
    assert regression_model.locked is True

    regression_model.unlock()
    assert regression_model.locked is False


def test_delete_model(scored_memoryset: ScoredMemoryset):
    RegressionModel.create("regression_model_to_delete", ScoredMemoryset.open(scored_memoryset.name))
    assert RegressionModel.open("regression_model_to_delete")
    RegressionModel.drop("regression_model_to_delete")
    with pytest.raises(LookupError):
        RegressionModel.open("regression_model_to_delete")


def test_delete_model_unauthenticated(unauthenticated_client, regression_model: RegressionModel):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            RegressionModel.drop(regression_model.name)


def test_delete_model_not_found():
    with pytest.raises(LookupError):
        RegressionModel.drop(str(uuid4()))
    # ignores error if specified
    RegressionModel.drop(str(uuid4()), if_not_exists="ignore")


def test_delete_model_unauthorized(unauthorized_client, regression_model: RegressionModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            RegressionModel.drop(regression_model.name)


def test_delete_memoryset_before_model_constraint_violation(datasource: Datasource):
    memoryset = ScoredMemoryset.from_datasource("test_memoryset_delete_before_regression_model", datasource=datasource)
    RegressionModel.create("test_regression_model_delete_before_memoryset", memoryset)
    with pytest.raises(RuntimeError):
        ScoredMemoryset.drop(memoryset.id)


def test_delete_memoryset_with_model_cascade(datasource: Datasource):
    """Test that cascade=False prevents deletion and cascade=True allows it."""
    memoryset = ScoredMemoryset.from_datasource("test_memoryset_cascade_delete_regression", datasource=datasource)
    model = RegressionModel.create("test_regression_model_cascade_delete", memoryset)

    # Verify model exists
    assert RegressionModel.open(model.name) is not None

    # Without cascade, deletion should fail
    with pytest.raises(RuntimeError):
        ScoredMemoryset.drop(memoryset.id, cascade=False)

    # Model should still exist
    assert RegressionModel.exists(model.name)

    # With cascade, deletion should succeed
    ScoredMemoryset.drop(memoryset.id, cascade=True)

    # Model should be deleted along with the memoryset
    assert not RegressionModel.exists(model.name)
    assert not ScoredMemoryset.exists(memoryset.name)


def test_evaluate_datasource(regression_model: RegressionModel, eval_datasource: Datasource):
    """Test that model evaluation with a datasource works."""
    result = regression_model.evaluate(eval_datasource)

    assert isinstance(result, RegressionMetrics)
    assert result.mae == pytest.approx(0.4)
    assert 0.0 <= result.mse <= 1.0
    assert 0.0 <= result.rmse <= 1.0
    assert result.r2 is not None

    assert isinstance(result.anomaly_score_mean, float)
    assert isinstance(result.anomaly_score_median, float)
    assert isinstance(result.anomaly_score_variance, float)
    assert -1.0 <= result.anomaly_score_mean <= 1.0
    assert -1.0 <= result.anomaly_score_median <= 1.0
    assert -1.0 <= result.anomaly_score_variance <= 1.0


def test_evaluate_local_dataset(regression_model: RegressionModel, eval_data: list[dict]):
    datasets = pytest.importorskip("datasets")
    eval_dataset = datasets.Dataset.from_list(eval_data)
    result = regression_model.evaluate(eval_dataset)

    assert isinstance(result, RegressionMetrics)
    assert isinstance(result.mae, float)
    assert isinstance(result.mse, float)
    assert isinstance(result.rmse, float)
    assert result.r2 is not None


def test_evaluate_local_list(regression_model: RegressionModel, eval_data: list[dict]):
    result = regression_model.evaluate(eval_data)
    assert result is not None
    assert isinstance(result, RegressionMetrics)
    assert isinstance(result.mae, float)
    assert isinstance(result.mse, float)
    assert isinstance(result.rmse, float)


def test_evaluate_datasource_with_nones_raises_error(regression_model: RegressionModel, datasource: Datasource):
    with pytest.raises(ValueError):
        regression_model.evaluate(datasource, record_predictions=True, tags={"test"})


def test_evaluate_with_telemetry(regression_model, eval_data: list[dict], eval_datasource: Datasource):
    result = regression_model.evaluate(eval_datasource, record_predictions=True, tags={"test"}, batch_size=2)
    assert result is not None
    assert isinstance(result, RegressionMetrics)
    predictions = regression_model.predictions(tag="test", batch_size=100, sort=[("timestamp", "asc")])
    assert len(predictions) == 4
    assert all(p.tags == {"test"} for p in predictions)
    assert all(p.expected_score is not None for p in predictions)
    prediction_expected_scores = [p.expected_score for p in predictions]
    eval_expected_scores = [d["score"] for d in eval_data]
    assert prediction_expected_scores == pytest.approx(eval_expected_scores, abs=1e-3)


def test_evaluate_with_partition_column_include_global(partitioned_regression_model: RegressionModel):
    result = partitioned_regression_model.evaluate(
        [
            {"value": "soup is good", "score": 0.1, "partition_id": "p1"},
            {"value": "cats are cute", "score": 0.9, "partition_id": "p1"},
        ],
        partition_column="partition_id",
        partition_filter_mode="include_global",
    )
    assert result is not None
    assert isinstance(result, RegressionMetrics)


def test_evaluate_with_partition_column_exclude_global(partitioned_regression_model: RegressionModel):
    result = partitioned_regression_model.evaluate(
        [
            {"value": "soup is good", "score": 0.1, "partition_id": "p1"},
            {"value": "cats are cute", "score": 0.9, "partition_id": "p1"},
        ],
        partition_column="partition_id",
        partition_filter_mode="exclude_global",
    )
    assert result is not None
    assert isinstance(result, RegressionMetrics)


def test_evaluate_with_partition_column_only_global(partitioned_regression_model: RegressionModel):
    result = partitioned_regression_model.evaluate(
        [
            {"value": "cats are independent animals", "score": 0.9, "partition_id": None},
            {"value": "i love the beach", "score": 0.5, "partition_id": None},
        ],
        partition_column="partition_id",
        partition_filter_mode="only_global",
    )
    assert result is not None
    assert isinstance(result, RegressionMetrics)


def test_evaluate_with_partition_column_ignore_partitions(partitioned_regression_model: RegressionModel):
    result = partitioned_regression_model.evaluate(
        [
            {"value": "soup is good", "score": 0.1, "partition_id": "p1"},
            {"value": "cats are cute", "score": 0.9, "partition_id": "p2"},
        ],
        partition_column="partition_id",
        partition_filter_mode="ignore_partitions",
    )
    assert result is not None
    assert isinstance(result, RegressionMetrics)


def test_evaluate_with_partition_column_datasource(partitioned_regression_model: RegressionModel):
    eval_datasource = Datasource.from_list(
        "eval_partition_datasource_reg",
        [
            {"value": "soup is good", "score": 0.1, "partition_id": "p1"},
            {"value": "cats are cute", "score": 0.9, "partition_id": "p1"},
            {"value": "homemade soup recipes", "score": 0.1, "partition_id": "p2"},
            {"value": "cats purr when happy", "score": 0.9, "partition_id": "p2"},
        ],
    )
    result = partitioned_regression_model.evaluate(
        eval_datasource,
        partition_column="partition_id",
        partition_filter_mode="exclude_global",
    )
    assert result is not None
    assert isinstance(result, RegressionMetrics)


def test_predict(regression_model: RegressionModel):
    predictions = regression_model.predict(["Do you love soup?", "Are cats cute?"], batch_size=1)
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    assert predictions[0].score is not None
    assert predictions[1].score is not None
    assert predictions[0].score == pytest.approx(0.1)
    assert predictions[1].score == pytest.approx(0.9)
    assert 0 <= predictions[0].confidence <= 1
    assert 0 <= predictions[1].confidence <= 1


def test_regression_prediction_has_no_score(regression_model: RegressionModel):
    """Ensure optional score is None for regression predictions."""
    prediction = regression_model.predict("This beach is amazing!")
    assert isinstance(prediction, RegressionPrediction)
    assert prediction.score is None


def test_predict_unauthenticated(unauthenticated_client, regression_model: RegressionModel):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            regression_model.predict(["This is excellent!", "This is terrible!"])


def test_predict_unauthorized(unauthorized_client, regression_model: RegressionModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            regression_model.predict(["This is excellent!", "This is terrible!"])


def test_predict_constraint_violation(scored_memoryset: ScoredMemoryset):
    model = RegressionModel.create(
        "test_regression_model_lookup_count_too_high",
        scored_memoryset,
        memory_lookup_count=scored_memoryset.length + 2,
    )
    with pytest.raises(RuntimeError):
        model.predict("test")


def test_predict_with_prompt(regression_model: RegressionModel):
    """Test that prompt parameter is properly passed through to predictions"""
    # Test with an instruction-supporting embedding model if available
    prediction_with_prompt = regression_model.predict(
        "This product is amazing!", prompt="Represent this text for rating prediction:"
    )
    prediction_without_prompt = regression_model.predict("This product is amazing!")

    # Both should work and return valid predictions
    assert prediction_with_prompt.score is not None
    assert prediction_without_prompt.score is not None
    assert 0 <= prediction_with_prompt.confidence <= 1
    assert 0 <= prediction_without_prompt.confidence <= 1


def test_predict_with_partition_id(partitioned_regression_model: RegressionModel):
    """Test predict with a specific partition_id"""
    # Predict with partition_id p1 - should use memories from p1
    prediction = partitioned_regression_model.predict("soup", partition_id="p1", partition_filter_mode="exclude_global")
    assert prediction.score is not None
    assert 0 <= prediction.confidence <= 1

    # Predict with partition_id p2 - should use memories from p2
    prediction_p2 = partitioned_regression_model.predict(
        "cats", partition_id="p2", partition_filter_mode="exclude_global"
    )
    assert prediction_p2.score is not None
    assert 0 <= prediction_p2.confidence <= 1


def test_predict_with_partition_id_include_global(partitioned_regression_model: RegressionModel):
    """Test predict with partition_id and include_global mode (default)"""
    # Predict with partition_id p1 and include_global (default) - should include both p1 and global memories
    prediction = partitioned_regression_model.predict("soup", partition_id="p1", partition_filter_mode="include_global")
    assert prediction.score is not None
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_exclude_global(partitioned_regression_model: RegressionModel):
    """Test predict with partition_id and exclude_global mode"""
    # Predict with partition_id p1 and exclude_global - should only use p1 memories
    prediction = partitioned_regression_model.predict("soup", partition_id="p1", partition_filter_mode="exclude_global")
    assert prediction.score is not None
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_only_global(partitioned_regression_model: RegressionModel):
    """Test predict with partition_filter_mode only_global"""
    # Predict with only_global mode - should only use global memories
    prediction = partitioned_regression_model.predict("cats", partition_filter_mode="only_global")
    assert prediction.score is not None
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_ignore_partitions(partitioned_regression_model: RegressionModel):
    """Test predict with partition_filter_mode ignore_partitions"""
    # Predict with ignore_partitions mode - should ignore partition filtering
    prediction = partitioned_regression_model.predict("soup", partition_filter_mode="ignore_partitions")
    assert prediction.score is not None
    assert 0 <= prediction.confidence <= 1


def test_predict_batch_with_partition_id(partitioned_regression_model: RegressionModel):
    """Test batch predict with partition_id"""
    # Batch predict with partition_id p1
    predictions = partitioned_regression_model.predict(
        ["soup is good", "cats are cute"],
        partition_id="p1",
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.score is not None for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)


def test_predict_batch_with_list_of_partition_ids(partitioned_regression_model: RegressionModel):
    """Test batch predict with a list of partition_ids (one for each query input)"""
    # Batch predict with a list of partition_ids - one for each input
    # First input uses p1, second input uses p2
    predictions = partitioned_regression_model.predict(
        ["soup is good", "cats are cute"],
        partition_id=["p1", "p2"],
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.score is not None for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)

    # Verify that predictions were made using the correct partitions
    # Each prediction should use memories from its respective partition
    assert predictions[0].input_value == "soup is good"
    assert predictions[1].input_value == "cats are cute"


def test_record_prediction_feedback(regression_model: RegressionModel):
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    expected_scores = [0.9, 0.1]
    regression_model.record_feedback(
        {
            "prediction_id": p.prediction_id,
            "category": "accurate",
            "value": abs(p.score - expected_score) < 0.2 if p.score is not None else False,
        }
        for expected_score, p in zip(expected_scores, predictions)
    )


def test_record_prediction_feedback_missing_category(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!")
    with pytest.raises(ValueError):
        regression_model.record_feedback({"prediction_id": prediction.prediction_id, "value": True})


def test_record_prediction_feedback_invalid_value(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!")
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        regression_model.record_feedback(
            {"prediction_id": prediction.prediction_id, "category": "accurate", "value": "invalid"}
        )


def test_record_prediction_feedback_invalid_prediction_id(regression_model: RegressionModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        regression_model.record_feedback({"prediction_id": "invalid", "category": "accurate", "value": True})


def test_predict_with_memoryset_override(regression_model: RegressionModel, data: list[dict]):
    # Create a memoryset with different scores (inverted)
    inverted_scored_memoryset = ScoredMemoryset.from_list(
        "test_memoryset_inverted_scores",
        [{"value": x["value"], "score": (2.0 - x["score"]) if x["score"] is not None else None} for x in data],
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
    )
    original_predictions = regression_model.predict(["This is excellent!", "This is terrible!"])

    with regression_model.use_memoryset(inverted_scored_memoryset):
        override_predictions = regression_model.predict(
            ["This is excellent!", "This is terrible!"], tags={"override_test"}
        )
        # With inverted scores, the predictions should be different
        assert original_predictions[0].score is not None
        assert original_predictions[1].score is not None
        assert override_predictions[0].score is not None
        assert override_predictions[1].score is not None
        assert abs(override_predictions[0].score - original_predictions[0].score) > 0.1
        assert abs(override_predictions[1].score - original_predictions[1].score) > 0.1
        assert override_predictions[0].memoryset == inverted_scored_memoryset
        assert override_predictions[1].memoryset == inverted_scored_memoryset

    # After exiting context, predictions should be back to normal
    new_predictions = regression_model.predict(["This is excellent!", "This is terrible!"], tags={"no_override_test"})
    assert new_predictions[0].score is not None
    assert new_predictions[1].score is not None
    assert original_predictions[0].score is not None
    assert original_predictions[1].score is not None
    assert abs(new_predictions[0].score - original_predictions[0].score) < 0.1
    assert abs(new_predictions[1].score - original_predictions[1].score) < 0.1
    assert new_predictions[0].memoryset == regression_model.memoryset
    assert new_predictions[1].memoryset == regression_model.memoryset

    # Verify historical predictions have the correct memoryset from telemetry
    fetched_override_predictions = regression_model.predictions(tag="override_test")
    assert len(fetched_override_predictions) == 2
    assert fetched_override_predictions[0].memoryset == inverted_scored_memoryset
    assert fetched_override_predictions[1].memoryset == inverted_scored_memoryset

    fetched_no_override_predictions = regression_model.predictions(tag="no_override_test")
    assert len(fetched_no_override_predictions) == 2
    assert fetched_no_override_predictions[0].memoryset == regression_model.memoryset
    assert fetched_no_override_predictions[1].memoryset == regression_model.memoryset


def test_predict_with_expected_scores(regression_model: RegressionModel):
    prediction = regression_model.predict("This is excellent!", expected_scores=0.9)
    assert prediction.expected_score == 0.9


def test_regression_prediction_update(regression_model: RegressionModel):
    prediction = regression_model.predict("Test input", expected_scores=3.5)
    assert prediction.expected_score == 3.5
    assert prediction.tags == set()

    # Update expected score
    prediction.update(expected_score=4.5)
    assert prediction.expected_score == 4.5

    # Add tags
    prediction.update(tags={"test", "updated"})
    assert prediction.tags == {"test", "updated"}

    # Clear both
    prediction.update(expected_score=None, tags=None)
    assert prediction.expected_score is None
    assert prediction.tags == set()


def test_last_prediction_with_batch(regression_model: RegressionModel):
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    assert regression_model.last_prediction is not None
    assert regression_model.last_prediction.prediction_id == predictions[-1].prediction_id
    assert regression_model.last_prediction.input_value == "This is terrible!"
    assert regression_model._last_prediction_was_batch is True


def test_last_prediction_with_single(regression_model: RegressionModel):
    # Test that last_prediction is updated correctly with single prediction
    prediction = regression_model.predict("This is excellent!")
    assert regression_model.last_prediction is not None
    assert regression_model.last_prediction.prediction_id == prediction.prediction_id
    assert regression_model.last_prediction.input_value == "This is excellent!"
    assert regression_model._last_prediction_was_batch is False


def test_batch_predict(regression_model: RegressionModel):
    """Test batch predictions"""
    predictions = regression_model.predict(["test input 1", "test input 2", "test input 3"])
    assert len(predictions) == 3
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)


def test_batch_predict_with_expected_scores(regression_model: RegressionModel):
    """Test batch predictions with expected scores"""
    predictions = regression_model.predict(["input 1", "input 2"], expected_scores=[0.5, 0.8])
    assert len(predictions) == 2
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)


def test_use_memoryset(regression_model: RegressionModel, scored_memoryset: ScoredMemoryset):
    # Test that predictions work with a memoryset
    predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
    assert len(predictions) == 2
    assert all(isinstance(pred, RegressionPrediction) for pred in predictions)
    assert all(0 <= pred.confidence <= 1 for pred in predictions)

    # Test that predictions work with a different memoryset
    with regression_model.use_memoryset(scored_memoryset):
        predictions = regression_model.predict(["This is excellent!", "This is terrible!"])
        assert len(predictions) == 2
        assert all(isinstance(pred, RegressionPrediction) for pred in predictions)
        assert all(0 <= pred.confidence <= 1 for pred in predictions)


def test_drop(regression_model):
    """Test that model drop works."""
    name = regression_model.name
    RegressionModel.drop(name)
    assert not RegressionModel.exists(name)


def test_predict_with_empty_partition(fully_partitioned_regression_resources):
    datasource, memoryset, regression_model = fully_partitioned_regression_resources

    assert memoryset.length == 15

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        regression_model.predict("i love cats", partition_filter_mode="only_global")

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        regression_model.predict("i love cats", partition_filter_mode="exclude_global", partition_id="p_does_not_exist")

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        regression_model.evaluate(datasource, partition_filter_mode="only_global")


def test_regression_model_repr(regression_model: RegressionModel):
    assert repr(regression_model) == dedent(
        """\
        RegressionModel({
            name: 'test_regression_model',
            memory_lookup_count: 3,
            memoryset: ScoredMemoryset({
                name: 'test_scored_memoryset',
                length: 22,
                embedding_model: PretrainedEmbeddingModel({ name: GTE_BASE, embedding_dim: 768, max_seq_length: 8192, num_params: 137M }),
            }),
        })"""
    )


def test_regression_metrics_repr():
    metrics = RegressionMetrics(
        {
            "coverage": 1.0,
            "mse": 0.1600,
            "rmse": 0.4,
            "mae": 0.4,
            "r2": -0.3333,
            "explained_variance": 0.0,
            "loss": 0.16,
            "anomaly_score_mean": -0.1234,
            "anomaly_score_median": -0.1,
            "anomaly_score_variance": 0.0567,
        }
    )
    assert repr(metrics) == dedent(
        """\
        RegressionMetrics({
            mae: 0.4000,
            rmse: 0.4000,
            r2: -0.3333,
            anomaly_score: -0.1234 ± 0.0567,
        })"""
    )


def test_regression_metrics_repr_without_optional_fields():
    metrics = RegressionMetrics(
        {
            "coverage": 1.0,
            "mse": 0.16,
            "rmse": 0.4,
            "mae": 0.4,
            "r2": 0.75,
            "explained_variance": 0.5,
            "loss": 0.16,
        }
    )
    assert repr(metrics) == dedent(
        """\
        RegressionMetrics({
            mae: 0.4000,
            rmse: 0.4000,
            r2: 0.7500,
        })"""
    )


def test_regression_prediction_repr():
    prediction = RegressionPrediction(
        None,
        label=None,
        label_name=None,
        score=0.1567,
        confidence=0.9234,
        anomaly_score=-0.0567,
        memoryset=MagicMock(),
        model=MagicMock(),
        input_value="Do you love soup?",
    )
    assert repr(prediction) == (
        "RegressionPrediction({ score: 0.16, confidence: 0.92, anomaly_score: -0.06, input_value: 'Do you love soup?' })"
    )



class TqdmFileReader:
    def __init__(self, file_obj, pbar):
        self.file_obj = file_obj
        self.pbar = pbar

    def read(self, size=-1):
        data = self.file_obj.read(size)
        self.pbar.update(len(data))
        return data

    def __getattr__(self, attr):
        return getattr(self.file_obj, attr)



from collections import namedtuple
from dataclasses import dataclass

import pytest

from .torch_parsing import list_from_torch

pytest.importorskip("torch")

from torch.utils.data import DataLoader as TorchDataLoader  # noqa: E402
from torch.utils.data import Dataset as TorchDataset  # noqa: E402


def test_list_from_torch_dict_dataset(data: list[dict]):
    class PytorchDictDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return self.data[i]

        def __len__(self):
            return len(self.data)

    dataset = PytorchDictDataset()
    data_list = list_from_torch(dataset)

    assert isinstance(data_list, list)
    assert len(data_list) == len(dataset)
    assert set(list(data_list[0].keys())) == {"value", "label", "key", "score", "source_id", "partition_id"}


def test_list_from_torch_dataloader(data: list[dict]):
    class PytorchDictDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return self.data[i]

        def __len__(self):
            return len(self.data)

    dataset = PytorchDictDataset()

    def collate_fn(x: list[dict]):
        return {"value": [item["value"] for item in x], "label": [item["label"] for item in x]}

    dataloader = TorchDataLoader(dataset, batch_size=3, collate_fn=collate_fn)
    data_list = list_from_torch(dataloader)

    assert isinstance(data_list, list)
    assert len(data_list) == len(dataset)
    assert list(data_list[0].keys()) == ["value", "label"]


def test_list_from_torch_tuple_dataset(data: list[dict]):
    class PytorchTupleDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return self.data[i]["value"], self.data[i]["label"]

        def __len__(self):
            return len(self.data)

    dataset = PytorchTupleDataset()

    # raises error if no column names are passed in
    with pytest.raises(ValueError):
        list_from_torch(dataset)

    # raises error if not enough column names are passed in
    with pytest.raises(ValueError):
        list_from_torch(dataset, column_names=["value"])

    # creates list if correct number of column names are passed in
    data_list = list_from_torch(dataset, column_names=["value", "label"])
    assert isinstance(data_list, list)
    assert len(data_list) == len(dataset)
    assert list(data_list[0].keys()) == ["value", "label"]


def test_list_from_torch_named_tuple_dataset(data: list[dict]):
    # Given a Pytorch dataset that returns a namedtuple for each item
    DatasetTuple = namedtuple("DatasetTuple", ["value", "label"])

    class PytorchNamedTupleDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return DatasetTuple(self.data[i]["value"], self.data[i]["label"])

        def __len__(self):
            return len(self.data)

    dataset = PytorchNamedTupleDataset()
    data_list = list_from_torch(dataset)
    assert isinstance(data_list, list)
    assert len(data_list) == len(dataset)
    assert list(data_list[0].keys()) == ["value", "label"]


def test_list_from_torch_dataclass_dataset(data: list[dict]):
    @dataclass
    class DatasetItem:
        text: str
        label: int

    class PytorchDataclassDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return DatasetItem(text=self.data[i]["value"], label=self.data[i]["label"])

        def __len__(self):
            return len(self.data)

    dataset = PytorchDataclassDataset()
    data_list = list_from_torch(dataset)
    assert isinstance(data_list, list)
    assert len(data_list) == len(dataset)
    assert list(data_list[0].keys()) == ["text", "label"]


def test_list_from_torch_invalid_dataset(data: list[dict]):
    class PytorchInvalidDataset(TorchDataset):
        def __init__(self):
            self.data = data

        def __getitem__(self, i):
            return [self.data[i]["value"], self.data[i]["label"]]

        def __len__(self):
            return len(self.data)

    dataset = PytorchInvalidDataset()
    with pytest.raises(ValueError):
        list_from_torch(dataset)



from textwrap import dedent
from unittest.mock import MagicMock
from uuid import uuid4

import pytest

from .classification_model import ClassificationMetrics, ClassificationModel
from .datasource import Datasource
from .embedding_model import PretrainedEmbeddingModel
from .memoryset import LabeledMemoryset
from .telemetry import ClassificationPrediction


def test_create_model(classification_model: ClassificationModel, readonly_memoryset: LabeledMemoryset):
    assert classification_model is not None
    assert classification_model.name == "test_classification_model"
    assert classification_model.memoryset == readonly_memoryset
    assert classification_model.num_classes == 2
    assert classification_model.memory_lookup_count == 3


def test_create_model_already_exists_error(readonly_memoryset, classification_model):
    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset)
    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="error")


def test_create_model_already_exists_return(readonly_memoryset, classification_model):
    with pytest.raises(ValueError):
        ClassificationModel.create(
            "test_classification_model", readonly_memoryset, if_exists="open", balance_classes=True
        )

    with pytest.raises(ValueError):
        ClassificationModel.create(
            "test_classification_model", readonly_memoryset, if_exists="open", memory_lookup_count=37
        )

    with pytest.raises(ValueError):
        ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="open", num_classes=19)

    with pytest.raises(ValueError):
        ClassificationModel.create(
            "test_classification_model", readonly_memoryset, if_exists="open", min_memory_weight=0.77
        )

    new_model = ClassificationModel.create("test_classification_model", readonly_memoryset, if_exists="open")
    assert new_model is not None
    assert new_model.name == "test_classification_model"
    assert new_model.memoryset == readonly_memoryset
    assert new_model.num_classes == 2
    assert new_model.memory_lookup_count == 3


def test_create_model_unauthenticated(unauthenticated_client, readonly_memoryset: LabeledMemoryset):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            ClassificationModel.create("test_model", readonly_memoryset)


def test_get_model(classification_model: ClassificationModel):
    fetched_model = ClassificationModel.open(classification_model.name)
    assert fetched_model is not None
    assert fetched_model.id == classification_model.id
    assert fetched_model.name == classification_model.name
    assert fetched_model.num_classes == 2
    assert fetched_model.memory_lookup_count == 3
    assert fetched_model == classification_model


def test_get_model_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            ClassificationModel.open("test_model")


def test_get_model_invalid_input():
    with pytest.raises(ValueError, match="Invalid input"):
        ClassificationModel.open("not valid id")


def test_get_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.open(str(uuid4()))


def test_get_model_unauthorized(unauthorized_client, classification_model: ClassificationModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            ClassificationModel.open(classification_model.name)


def test_list_models(classification_model: ClassificationModel):
    models = ClassificationModel.all()
    assert len(models) > 0
    assert any(model.name == model.name for model in models)


def test_list_models_unauthenticated(unauthenticated_client):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            ClassificationModel.all()


def test_list_models_unauthorized(unauthorized_client, classification_model: ClassificationModel):
    with unauthorized_client.use():
        assert ClassificationModel.all() == []


def test_memoryset_classification_models_property(
    classification_model: ClassificationModel, readonly_memoryset: LabeledMemoryset
):
    models = readonly_memoryset.classification_models
    assert len(models) > 0
    assert any(model.id == classification_model.id for model in models)


def test_update_model_attributes(classification_model: ClassificationModel):
    original_description = classification_model.description

    classification_model.set(description=None)
    assert classification_model.description is None

    classification_model.set(description=original_description)
    assert classification_model.description == original_description

    classification_model.set(locked=True)
    assert classification_model.locked is True

    classification_model.set(locked=False)
    assert classification_model.locked is False

    classification_model.lock()
    assert classification_model.locked is True

    classification_model.unlock()
    assert classification_model.locked is False


def test_delete_model(readonly_memoryset: LabeledMemoryset):
    ClassificationModel.create("model_to_delete", LabeledMemoryset.open(readonly_memoryset.name))
    assert ClassificationModel.open("model_to_delete")
    ClassificationModel.drop("model_to_delete")
    with pytest.raises(LookupError):
        ClassificationModel.open("model_to_delete")


def test_delete_model_unauthenticated(unauthenticated_client, classification_model: ClassificationModel):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            ClassificationModel.drop(classification_model.name)


def test_delete_model_not_found():
    with pytest.raises(LookupError):
        ClassificationModel.drop(str(uuid4()))
    # ignores error if specified
    ClassificationModel.drop(str(uuid4()), if_not_exists="ignore")


def test_delete_model_unauthorized(unauthorized_client, classification_model: ClassificationModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            ClassificationModel.drop(classification_model.name)


def test_delete_memoryset_before_model_constraint_violation(datasource):
    memoryset = LabeledMemoryset.from_datasource(
        "test_memoryset_delete_before_model", datasource=datasource, label_names=["soup", "cats"]
    )
    ClassificationModel.create("test_model_delete_before_memoryset", memoryset)
    with pytest.raises(RuntimeError):
        LabeledMemoryset.drop(memoryset.id)


def test_delete_memoryset_with_model_cascade(datasource):
    """Test that cascade=False prevents deletion and cascade=True allows it."""
    memoryset = LabeledMemoryset.from_datasource(
        "test_memoryset_cascade_delete", datasource=datasource, label_names=["soup", "cats"]
    )
    model = ClassificationModel.create("test_model_cascade_delete", memoryset)

    # Verify model exists
    assert ClassificationModel.open(model.name) is not None

    # Without cascade, deletion should fail
    with pytest.raises(RuntimeError):
        LabeledMemoryset.drop(memoryset.id, cascade=False)

    # Model should still exist
    assert ClassificationModel.exists(model.name)

    # With cascade, deletion should succeed
    LabeledMemoryset.drop(memoryset.id, cascade=True)

    # Model should be deleted along with the memoryset
    assert not ClassificationModel.exists(model.name)
    assert not LabeledMemoryset.exists(memoryset.name)


def test_evaluate(classification_model, eval_datasource: Datasource):
    result = classification_model.evaluate(eval_datasource)

    assert result is not None
    assert isinstance(result, ClassificationMetrics)

    assert isinstance(result.accuracy, float)
    assert result.accuracy == pytest.approx(0.5)
    assert isinstance(result.f1_score, float)
    assert result.f1_score == pytest.approx(0.5)
    assert isinstance(result.loss, float)

    assert isinstance(result.anomaly_score_mean, float)
    assert isinstance(result.anomaly_score_median, float)
    assert isinstance(result.anomaly_score_variance, float)
    assert -1.0 <= result.anomaly_score_mean <= 1.0
    assert -1.0 <= result.anomaly_score_median <= 1.0
    assert -1.0 <= result.anomaly_score_variance <= 1.0

    assert result.pr_auc is not None
    assert result.pr_auc == pytest.approx(0.83333, rel=1e-5)
    assert result.pr_curve is not None
    assert result.pr_curve["thresholds"] == pytest.approx(
        [0.0, 0.3021204173564911, 0.30852025747299194, 0.6932827234268188, 0.6972201466560364]
    )
    assert result.pr_curve["precisions"] == pytest.approx([0.5, 2 / 3, 0.5, 1.0, 1.0])
    assert result.pr_curve["recalls"] == pytest.approx([1.0, 1.0, 0.5, 0.5, 0.0])

    assert result.roc_auc is not None
    assert result.roc_auc == pytest.approx(0.75)
    assert result.roc_curve is not None
    assert result.roc_curve["thresholds"] == pytest.approx(
        [0.3021204173564911, 0.30852025747299194, 0.6932827234268188, 0.6972201466560364, 1.0]
    )
    assert result.roc_curve["false_positive_rates"] == pytest.approx([1.0, 0.5, 0.5, 0.0, 0.0])
    assert result.roc_curve["true_positive_rates"] == pytest.approx([1.0, 1.0, 0.5, 0.5, 0.0])


def test_evaluate_local_dataset(classification_model, eval_data: list[dict]):
    datasets = pytest.importorskip("datasets")
    eval_dataset = datasets.Dataset.from_list(eval_data)
    result = classification_model.evaluate(eval_dataset)
    assert result is not None
    assert isinstance(result, ClassificationMetrics)
    assert isinstance(result.accuracy, float)
    assert isinstance(result.f1_score, float)
    assert isinstance(result.loss, float)


def test_evaluate_local_list(classification_model, eval_data: list[dict]):
    result = classification_model.evaluate(eval_data)
    assert result is not None
    assert isinstance(result, ClassificationMetrics)
    assert isinstance(result.accuracy, float)
    assert isinstance(result.f1_score, float)
    assert isinstance(result.loss, float)


def test_evaluate_datasource_with_nones_raises_error(classification_model: ClassificationModel, datasource: Datasource):
    with pytest.raises(ValueError):
        classification_model.evaluate(datasource, record_predictions=True, tags={"test"})


def test_evaluate_with_telemetry(classification_model: ClassificationModel, eval_data: list[dict]):
    result = classification_model.evaluate(eval_data, record_predictions=True, tags={"test"}, batch_size=2)
    assert result is not None
    assert isinstance(result, ClassificationMetrics)
    predictions = classification_model.predictions(tag="test", batch_size=100, sort=[("timestamp", "asc")])
    assert len(predictions) == 4
    assert all(p.tags == {"test"} for p in predictions)
    prediction_expected_labels = [p.expected_label if p.expected_label is not None else -1 for p in predictions]
    eval_expected_labels = [d["label"] for d in eval_data]
    assert all(
        p == l for p, l in zip(prediction_expected_labels, eval_expected_labels)
    ), f"Prediction expected labels: {prediction_expected_labels} do not match eval expected labels: {eval_expected_labels}"


def test_evaluate_with_partition_column_include_global(partitioned_classification_model: ClassificationModel):
    result = partitioned_classification_model.evaluate(
        [
            {"value": "soup is good", "label": 0, "partition_id": "p1"},
            {"value": "cats are cute", "label": 1, "partition_id": "p1"},
        ],
        partition_column="partition_id",
        partition_filter_mode="include_global",
    )
    assert result is not None
    assert isinstance(result, ClassificationMetrics)


def test_evaluate_with_partition_column_exclude_global(partitioned_classification_model: ClassificationModel):
    result = partitioned_classification_model.evaluate(
        [
            {"value": "soup is good", "label": 0, "partition_id": "p1"},
            {"value": "cats are cute", "label": 1, "partition_id": "p1"},
        ],
        partition_column="partition_id",
        partition_filter_mode="exclude_global",
    )
    assert result is not None
    assert isinstance(result, ClassificationMetrics)


def test_evaluate_with_partition_column_only_global(partitioned_classification_model: ClassificationModel):
    result = partitioned_classification_model.evaluate(
        [
            {"value": "cats are independent animals", "label": 1, "partition_id": None},
            {"value": "i love the beach", "label": 1, "partition_id": None},
        ],
        partition_column="partition_id",
        partition_filter_mode="only_global",
    )
    assert result is not None
    assert isinstance(result, ClassificationMetrics)


def test_evaluate_with_partition_column_ignore_partitions(partitioned_classification_model: ClassificationModel):
    result = partitioned_classification_model.evaluate(
        [
            {"value": "soup is good", "label": 0, "partition_id": "p1"},
            {"value": "cats are cute", "label": 1, "partition_id": "p2"},
        ],
        partition_column="partition_id",
        partition_filter_mode="ignore_partitions",
    )
    assert result is not None
    assert isinstance(result, ClassificationMetrics)


def test_evaluate_with_partition_column_datasource(partitioned_classification_model: ClassificationModel):
    eval_datasource = Datasource.from_list(
        "eval_partition_datasource",
        [
            {"value": "soup is good", "label": 0, "partition_id": "p1"},
            {"value": "cats are cute", "label": 1, "partition_id": "p1"},
            {"value": "homemade soup recipes", "label": 0, "partition_id": "p2"},
            {"value": "cats purr when happy", "label": 1, "partition_id": "p2"},
        ],
    )
    result = partitioned_classification_model.evaluate(
        eval_datasource,
        partition_column="partition_id",
        partition_filter_mode="exclude_global",
    )
    assert result is not None
    assert isinstance(result, ClassificationMetrics)


def test_predict(classification_model: ClassificationModel, label_names: list[str]):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"], batch_size=1)
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1

    assert predictions[0].logits is not None
    assert predictions[1].logits is not None
    assert len(predictions[0].logits) == 2
    assert len(predictions[1].logits) == 2
    assert predictions[0].logits[0] > predictions[0].logits[1]
    assert predictions[1].logits[0] < predictions[1].logits[1]


def test_classification_prediction_has_no_label(classification_model: ClassificationModel):
    """Ensure optional score is None for classification predictions."""
    prediction = classification_model.predict("Do you want to go to the beach?")
    assert isinstance(prediction, ClassificationPrediction)
    assert prediction.label is None


def test_predict_disable_telemetry(classification_model: ClassificationModel, label_names: list[str]):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"], save_telemetry="off")
    assert len(predictions) == 2
    assert predictions[0].prediction_id is None
    assert predictions[1].prediction_id is None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1


def test_predict_unauthenticated(unauthenticated_client, classification_model: ClassificationModel):
    with unauthenticated_client.use():
        with pytest.raises(ValueError, match="Invalid API key"):
            classification_model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_unauthorized(unauthorized_client, classification_model: ClassificationModel):
    with unauthorized_client.use():
        with pytest.raises(LookupError):
            classification_model.predict(["Do you love soup?", "Are cats cute?"])


def test_predict_constraint_violation(readonly_memoryset: LabeledMemoryset):
    model = ClassificationModel.create(
        "test_model_lookup_count_too_high",
        readonly_memoryset,
        num_classes=2,
        memory_lookup_count=readonly_memoryset.length + 2,
    )
    with pytest.raises(RuntimeError):
        model.predict("test")


def test_predict_with_partition_id(partitioned_classification_model: ClassificationModel, label_names: list[str]):
    """Test predict with a specific partition_id"""
    # Predict with partition_id p1 - should use memories from p1
    prediction = partitioned_classification_model.predict(
        "soup", partition_id="p1", partition_filter_mode="exclude_global"
    )
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1
    assert prediction.logits is not None
    assert len(prediction.logits) == 2

    # Predict with partition_id p2 - should use memories from p2
    prediction_p2 = partitioned_classification_model.predict(
        "cats", partition_id="p2", partition_filter_mode="exclude_global"
    )
    assert prediction_p2.label is not None
    assert prediction_p2.label_name in label_names
    assert 0 <= prediction_p2.confidence <= 1


def test_predict_with_partition_id_include_global(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test predict with partition_id and include_global mode (default)"""
    # Predict with partition_id p1 and include_global (default) - should include both p1 and global memories
    prediction = partitioned_classification_model.predict(
        "soup", partition_id="p1", partition_filter_mode="include_global"
    )
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_exclude_global(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test predict with partition_id and exclude_global mode"""
    # Predict with partition_id p1 and exclude_global - should only use p1 memories
    prediction = partitioned_classification_model.predict(
        "soup", partition_id="p1", partition_filter_mode="exclude_global"
    )
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_only_global(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test predict with partition_filter_mode only_global"""
    # Predict with only_global mode - should only use global memories
    prediction = partitioned_classification_model.predict("cats", partition_filter_mode="only_global")
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1


def test_predict_with_partition_id_ignore_partitions(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test predict with partition_filter_mode ignore_partitions"""
    # Predict with ignore_partitions mode - should ignore partition filtering
    prediction = partitioned_classification_model.predict("soup", partition_filter_mode="ignore_partitions")
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1


def test_predict_batch_with_partition_id(partitioned_classification_model: ClassificationModel, label_names: list[str]):
    """Test batch predict with partition_id"""
    # Batch predict with partition_id p1
    predictions = partitioned_classification_model.predict(
        ["soup is good", "cats are cute"],
        partition_id="p1",
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.label is not None for p in predictions)
    assert all(p.label_name in label_names for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)
    assert all(p.logits is not None and len(p.logits) == 2 for p in predictions)


def test_predict_with_partition_id_and_filters(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test predict with partition_id and filters"""
    # Predict with partition_id and filters
    prediction = partitioned_classification_model.predict(
        "soup",
        partition_id="p1",
        partition_filter_mode="exclude_global",
        filters=[("key", "==", "g1")],
    )
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1


def test_predict_batch_with_list_of_partition_ids(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test batch predict with a list of partition_ids (one for each query input)"""
    # Batch predict with a list of partition_ids - one for each input
    # First input uses p1, second input uses p2
    predictions = partitioned_classification_model.predict(
        ["soup is good", "cats are cute"],
        partition_id=["p1", "p2"],
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.label is not None for p in predictions)
    assert all(p.label_name in label_names for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)
    assert all(p.logits is not None and len(p.logits) == 2 for p in predictions)

    # Verify that predictions were made using the correct partitions
    # Each prediction should use memories from its respective partition
    assert predictions[0].input_value == "soup is good"
    assert predictions[1].input_value == "cats are cute"


@pytest.mark.asyncio
async def test_predict_async_with_partition_id(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test async predict with partition_id"""
    # Async predict with partition_id p1
    prediction = await partitioned_classification_model.apredict(
        "soup", partition_id="p1", partition_filter_mode="exclude_global"
    )
    assert prediction.label is not None
    assert prediction.label_name in label_names
    assert 0 <= prediction.confidence <= 1
    assert prediction.logits is not None
    assert len(prediction.logits) == 2


@pytest.mark.asyncio
async def test_predict_async_batch_with_partition_id(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test async batch predict with partition_id"""
    # Async batch predict with partition_id p1
    predictions = await partitioned_classification_model.apredict(
        ["soup is good", "cats are cute"],
        partition_id="p1",
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.label is not None for p in predictions)
    assert all(p.label_name in label_names for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)


@pytest.mark.asyncio
async def test_predict_async_batch_with_list_of_partition_ids(
    partitioned_classification_model: ClassificationModel, label_names: list[str]
):
    """Test async batch predict with a list of partition_ids (one for each query input)"""
    # Async batch predict with a list of partition_ids - one for each input
    # First input uses p1, second input uses p2
    predictions = await partitioned_classification_model.apredict(
        ["soup is good", "cats are cute"],
        partition_id=["p1", "p2"],
        partition_filter_mode="exclude_global",
    )
    assert len(predictions) == 2
    assert all(p.label is not None for p in predictions)
    assert all(p.label_name in label_names for p in predictions)
    assert all(0 <= p.confidence <= 1 for p in predictions)
    assert all(p.logits is not None and len(p.logits) == 2 for p in predictions)

    # Verify that predictions were made using the correct partitions
    # Each prediction should use memories from its respective partition
    assert predictions[0].input_value == "soup is good"
    assert predictions[1].input_value == "cats are cute"


def test_record_prediction_feedback(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    expected_labels = [0, 1]
    classification_model.record_feedback(
        {
            "prediction_id": p.prediction_id,
            "category": "correct",
            "value": p.label == expected_label,
        }
        for expected_label, p in zip(expected_labels, predictions)
    )


def test_record_prediction_feedback_missing_category(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    with pytest.raises(ValueError):
        classification_model.record_feedback({"prediction_id": prediction.prediction_id, "value": True})


def test_record_prediction_feedback_invalid_value(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?")
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.record_feedback(
            {"prediction_id": prediction.prediction_id, "category": "correct", "value": "invalid"}
        )


def test_record_prediction_feedback_invalid_prediction_id(classification_model: ClassificationModel):
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.record_feedback({"prediction_id": "invalid", "category": "correct", "value": True})


def test_predict_with_memoryset_override(
    classification_model: ClassificationModel, data: list[dict], label_names: list[str]
):
    inverted_labeled_memoryset = LabeledMemoryset.from_list(
        "test_memoryset_inverted_labels",
        [{"value": x["value"], "label": 1 if x["label"] == 0 else 0} for x in data],
        label_names=label_names,
        embedding_model=PretrainedEmbeddingModel.GTE_BASE,
    )
    with classification_model.use_memoryset(inverted_labeled_memoryset):
        predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"], tags={"override_test"})
        assert predictions[0].label == 1
        assert predictions[1].label == 0
        assert predictions[0].memoryset == inverted_labeled_memoryset
        assert predictions[1].memoryset == inverted_labeled_memoryset

    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"], tags={"no_override_test"})
    assert predictions[0].label == 0
    assert predictions[1].label == 1
    assert predictions[0].memoryset == classification_model.memoryset
    assert predictions[1].memoryset == classification_model.memoryset

    # Verify historical predictions have the correct memoryset from telemetry
    override_predictions = classification_model.predictions(tag="override_test")
    assert len(override_predictions) == 2
    assert override_predictions[0].memoryset == inverted_labeled_memoryset
    assert override_predictions[1].memoryset == inverted_labeled_memoryset

    no_override_predictions = classification_model.predictions(tag="no_override_test")
    assert len(no_override_predictions) == 2
    assert no_override_predictions[0].memoryset == classification_model.memoryset
    assert no_override_predictions[1].memoryset == classification_model.memoryset


def test_predict_with_expected_labels(classification_model: ClassificationModel):
    prediction = classification_model.predict("Do you love soup?", expected_labels=1)
    assert prediction.expected_label == 1


def test_predict_with_expected_labels_no_telemetry(classification_model: ClassificationModel):
    """Test that expected_label is available even when telemetry is disabled"""
    prediction = classification_model.predict("Do you love soup?", expected_labels=1, save_telemetry="off")
    assert prediction.prediction_id is None  # telemetry is off
    assert prediction.expected_label == 1  # but expected_label should still be available


def test_predict_with_expected_labels_invalid_input(classification_model: ClassificationModel):
    # invalid number of expected labels for batch prediction
    with pytest.raises(ValueError, match=r"Invalid input.*"):
        classification_model.predict(["Do you love soup?", "Are cats cute?"], expected_labels=[0])
    # invalid label value
    with pytest.raises(ValueError):
        classification_model.predict("Do you love soup?", expected_labels=5)


def test_predict_with_filters(classification_model: ClassificationModel):
    # there are no memories with label 0 and key g1, so we force a wrong prediction
    filtered_prediction = classification_model.predict("I love soup", filters=[("key", "==", "g2")])
    assert filtered_prediction.label == 1
    assert filtered_prediction.label_name == "cats"


def test_predict_with_memoryset_update(writable_memoryset: LabeledMemoryset):
    model = ClassificationModel.create(
        "test_predict_with_memoryset_update",
        writable_memoryset,
        num_classes=2,
        memory_lookup_count=3,
    )
    try:
        prediction = model.predict("Do you love soup?", partition_filter_mode="ignore_partitions")
        assert prediction.label == 0
        assert prediction.label_name == "soup"
        # insert new memories
        writable_memoryset.insert(
            [
                {"value": "Do you love soup?", "label": 1, "key": "g1"},
                {"value": "Do you love soup for dinner?", "label": 1, "key": "g2"},
                {"value": "Do you love crackers?", "label": 1, "key": "g2"},
                {"value": "Do you love broth?", "label": 1, "key": "g2"},
                {"value": "Do you love chicken soup?", "label": 1, "key": "g2"},
                {"value": "Do you love chicken soup for dinner?", "label": 1, "key": "g2"},
                {"value": "Do you love chicken soup for dinner?", "label": 1, "key": "g2"},
            ],
        )
        prediction = model.predict("Do you love soup?")
        assert prediction.label == 1
        assert prediction.label_name == "cats"
    finally:
        ClassificationModel.drop("test_predict_with_memoryset_update")


def test_last_prediction_with_batch(classification_model: ClassificationModel):
    predictions = classification_model.predict(["Do you love soup?", "Are cats cute?"])
    assert classification_model.last_prediction is not None
    assert classification_model.last_prediction.prediction_id == predictions[-1].prediction_id
    assert classification_model.last_prediction.input_value == "Are cats cute?"
    assert classification_model._last_prediction_was_batch is True


def test_last_prediction_with_single(classification_model: ClassificationModel):
    # Test that last_prediction is updated correctly with single prediction
    prediction = classification_model.predict("Do you love soup?")
    assert classification_model.last_prediction is not None
    assert classification_model.last_prediction.prediction_id == prediction.prediction_id
    assert classification_model.last_prediction.input_value == "Do you love soup?"
    assert classification_model._last_prediction_was_batch is False


def test_explain(classification_model: ClassificationModel, requires_agentic_features):
    prediction = classification_model.predict("Do you love soup?")
    explanation = prediction.explanation
    assert explanation is not None
    assert len(explanation) > 10
    assert "soup" in explanation.lower()


def test_action_recommendation(
    classification_model: ClassificationModel, writable_memoryset: LabeledMemoryset, requires_agentic_features
):
    with classification_model.use_memoryset(writable_memoryset):
        prediction = classification_model.predict("Do you love soup?", expected_labels=1)
    assert prediction.memoryset == writable_memoryset

    memoryset_length = writable_memoryset.length

    # Get action recommendation
    action, rationale = prediction.recommend_action()

    assert action is not None
    assert rationale is not None
    assert action in ["remove_duplicates", "detect_mislabels", "add_memories", "finetuning"]
    assert len(rationale) > 10

    # Test memory suggestions
    suggestions = prediction.generate_memory_suggestions(num_memories=2)

    assert len(suggestions) == 2
    for suggestion in suggestions:
        assert isinstance(suggestion.value, str)
        assert len(suggestion.value) > 0
        assert isinstance(suggestion.label_name, str)
        assert suggestion.label_name in writable_memoryset.label_names

    writable_memoryset.insert(suggestions)

    writable_memoryset.refresh()
    assert writable_memoryset.length == memoryset_length + 2


def test_predict_with_prompt(classification_model: ClassificationModel):
    """Test that prompt parameter is properly passed through to predictions"""
    # Test with an instruction-supporting embedding model if available
    prediction_with_prompt = classification_model.predict(
        "I love this product!", prompt="Represent this text for sentiment classification:"
    )
    prediction_without_prompt = classification_model.predict("I love this product!")

    # Both should work and return valid predictions
    assert prediction_with_prompt.label is not None
    assert prediction_without_prompt.label is not None


def test_predict_with_empty_partition(fully_partitioned_classification_resources):
    datasource, memoryset, classification_model = fully_partitioned_classification_resources

    assert memoryset.length == 15

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        classification_model.predict("i love cats", partition_filter_mode="only_global")

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        classification_model.predict(
            "i love cats", partition_filter_mode="exclude_global", partition_id="p_does_not_exist"
        )

    with pytest.raises(RuntimeError, match="lookup failed to return the correct number of memories"):
        classification_model.evaluate(datasource, partition_filter_mode="only_global")


@pytest.mark.asyncio
async def test_predict_async_single(classification_model: ClassificationModel, label_names: list[str]):
    """Test async prediction with a single value"""
    prediction = await classification_model.apredict("Do you love soup?")
    assert isinstance(prediction, ClassificationPrediction)
    assert prediction.prediction_id is not None
    assert prediction.label == 0
    assert prediction.label_name == label_names[0]
    assert 0 <= prediction.confidence <= 1
    assert prediction.logits is not None
    assert len(prediction.logits) == 2


@pytest.mark.asyncio
async def test_predict_async_batch(classification_model: ClassificationModel, label_names: list[str]):
    """Test async prediction with a batch of values"""
    predictions = await classification_model.apredict(["Do you love soup?", "Are cats cute?"])
    assert len(predictions) == 2
    assert predictions[0].prediction_id is not None
    assert predictions[1].prediction_id is not None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1


@pytest.mark.asyncio
async def test_predict_async_with_expected_labels(classification_model: ClassificationModel):
    """Test async prediction with expected labels"""
    prediction = await classification_model.apredict("Do you love soup?", expected_labels=1)
    assert prediction.expected_label == 1


@pytest.mark.asyncio
async def test_predict_async_disable_telemetry(classification_model: ClassificationModel, label_names: list[str]):
    """Test async prediction with telemetry disabled"""
    predictions = await classification_model.apredict(["Do you love soup?", "Are cats cute?"], save_telemetry="off")
    assert len(predictions) == 2
    assert predictions[0].prediction_id is None
    assert predictions[1].prediction_id is None
    assert predictions[0].label == 0
    assert predictions[0].label_name == label_names[0]
    assert 0 <= predictions[0].confidence <= 1
    assert predictions[1].label == 1
    assert predictions[1].label_name == label_names[1]
    assert 0 <= predictions[1].confidence <= 1


@pytest.mark.asyncio
async def test_predict_async_with_filters(classification_model: ClassificationModel):
    """Test async prediction with filters"""
    # there are no memories with label 0 and key g2, so we force a wrong prediction
    filtered_prediction = await classification_model.apredict("I love soup", filters=[("key", "==", "g2")])
    assert filtered_prediction.label == 1
    assert filtered_prediction.label_name == "cats"


def test_classification_model_repr(classification_model: ClassificationModel):
    assert repr(classification_model) == dedent(
        """\
        ClassificationModel({
            name: 'test_classification_model',
            num_classes: 2,
            memory_lookup_count: 3,
            weigh_memories: True,
            balance_classes: False,
            memoryset: LabeledMemoryset({
                name: 'test_readonly_memoryset',
                length: 22,
                embedding_model: PretrainedEmbeddingModel({ name: GTE_BASE, embedding_dim: 768, max_seq_length: 8192, num_params: 137M }),
            }),
        })"""
    )


def test_classification_metrics_repr():
    metrics = ClassificationMetrics(
        {
            "coverage": 1.0,
            "accuracy": 0.8333,
            "f1_score": 0.7567,
            "loss": 0.5,
            "anomaly_score_mean": -0.1234,
            "anomaly_score_median": -0.1,
            "anomaly_score_variance": 0.0567,
            "roc_auc": 0.9123,
            "pr_auc": 0.8765,
        }
    )
    assert repr(metrics) == dedent(
        """\
        ClassificationMetrics({
            accuracy: 0.8333,
            f1_score: 0.7567,
            roc_auc: 0.9123,
            pr_auc: 0.8765,
            anomaly_score: -0.1234 ± 0.0567,
        })"""
    )


def test_classification_metrics_repr_without_optional_fields():
    metrics = ClassificationMetrics(
        {
            "coverage": 1.0,
            "accuracy": 0.5,
            "f1_score": 0.5,
            "loss": None,
        }
    )
    assert repr(metrics) == dedent(
        """\
        ClassificationMetrics({
            accuracy: 0.5000,
            f1_score: 0.5000,
        })"""
    )


def test_classification_prediction_repr():
    prediction = ClassificationPrediction(
        None,
        label=0,
        label_name="soup",
        score=None,
        confidence=0.8567,
        anomaly_score=-0.1234,
        memoryset=MagicMock(),
        model=MagicMock(),
        input_value="Do you love soup?",
    )
    assert (
        repr(prediction)
        == "ClassificationPrediction({ label: <soup: 0>, confidence: 0.86, anomaly_score: -0.12, input_value: 'Do you love soup?' })"
    )



from datetime import datetime
from typing import Literal

import httpx
from httpx import ConnectError, Headers

from .async_client import OrcaAsyncClient
from .client import OrcaClient

Scope = Literal["ADMINISTER", "PREDICT"]
"""
The scopes of an API key.

- `ADMINISTER`: Can create any resource in the organization.
- `PREDICT`: Can only call model.predict and perform CRUD operations on predictions.
"""


class ApiKeyInfo:
    """
    Information about an API key

    Note:
        The value of the API key is only available at creation time.

    Attributes:
        name: Unique name of the API key
        created_at: When the API key was created
        scopes: The scopes of the API key
    """

    name: str
    created_at: datetime
    scopes: set[Scope]

    def __init__(self, name: str, created_at: datetime, scopes: set[Scope]):
        self.name = name
        self.created_at = created_at
        self.scopes = scopes

    def __repr__(self) -> str:
        return "ApiKey({ " + f"name: '{self.name}', scopes: <{'|'.join(self.scopes)}>" + " })"


class OrcaCredentials:
    """
    Class for managing Orca API credentials
    """

    @staticmethod
    def is_authenticated() -> bool:
        """
        Check if you are authenticated to interact with the Orca API

        Returns:
            True if you are authenticated, False otherwise
        """
        client = OrcaClient._resolve_client()
        try:
            return client.GET("/auth")
        except ValueError as e:
            if "Invalid API key" in str(e):
                return False
            raise e

    @staticmethod
    def is_healthy() -> bool:
        """
        Check whether the API is healthy

        Returns:
            True if the API is healthy, False otherwise
        """
        client = OrcaClient._resolve_client()
        try:
            # use httpx directly to avoid retries, with a short timeout to fail fast
            base_url = str(client.base_url).rstrip("/")
            response = httpx.get(f"{base_url}/check/healthy", timeout=5)
            return response.status_code == 200 and response.json().get("ok", False)
        except Exception:
            return False

    @staticmethod
    def get_config() -> dict[str, str | float | int | bool | None]:
        """
        Get the config of the Orca API

        Returns:
            A dictionary of the public config of the Orca API
        """
        client = OrcaClient._resolve_client()
        return client.GET("/config")

    @staticmethod
    def get_org_id() -> str:
        """
        Get the organization ID of the currently authenticated user

        Returns:
            The organization ID
        """
        client = OrcaClient._resolve_client()
        return client.GET("/auth/info")["org_id"]

    @staticmethod
    def get_api_key_scopes() -> set[Scope | Literal["ROOT"]]:
        """
        Get the scopes of the API key that is currently being used to authenticate with the Orca API

        Returns:
            A set of scopes
        """
        client = OrcaClient._resolve_client()
        res = client.GET("/auth/info")
        if res["is_root"]:
            return set(["ROOT", "ADMINISTER"])
        return set(res["scope"])

    @staticmethod
    def get_api_key_name() -> str:
        """
        Get the name of the API key that is currently being used to authenticate with the Orca API

        Returns:
            The name of the API key
        """
        client = OrcaClient._resolve_client()
        return client.GET("/auth/info")["api_key_name"] or "ROOT_ACCESS_API_KEY"

    @staticmethod
    def list_api_keys() -> list[ApiKeyInfo]:
        """
        List all API keys that have been created for your org

        Returns:
            A list of named tuples, with the name and creation date time of the API key
        """
        client = OrcaClient._resolve_client()
        return [
            ApiKeyInfo(
                name=api_key["name"],
                created_at=datetime.fromisoformat(api_key["created_at"]),
                scopes=set(api_key["scope"]),
            )
            for api_key in client.GET("/auth/api_key")
        ]

    @staticmethod
    def create_api_key(name: str, scopes: set[Scope] = {"ADMINISTER"}) -> str:
        """
        Create a new API key with the given name and scopes

        Params:
            name: The name of the API key
            scopes: The scopes of the API key

        Returns:
            The secret value of the API key. Make sure to save this value as it will not be shown again.
        """
        client = OrcaClient._resolve_client()
        res = client.POST(
            "/auth/api_key",
            json={"name": name, "scope": list(scopes)},
        )
        return res["api_key"]

    @staticmethod
    def revoke_api_key(name: str) -> None:
        """
        Delete an API key

        Params:
            name: The name of the API key to delete

        Raises:
            ValueError: if the API key is not found
        """
        client = OrcaClient._resolve_client()
        client.DELETE("/auth/api_key/{name_or_id}", params={"name_or_id": name})

    @staticmethod
    def set_api_key(api_key: str, check_validity: bool = True):
        """
        Set the API key to use for authenticating with the Orca API

        Note:
            The API key can also be provided by setting the `ORCA_API_KEY` environment variable

        Params:
            api_key: The API key to set
            check_validity: Whether to check if the API key is valid and raise an error otherwise

        Raises:
            ValueError: if the API key is invalid and `check_validity` is True
        """
        sync_client = OrcaClient._resolve_client()
        sync_client.api_key = api_key
        if check_validity:
            sync_client.GET("/auth")

        async_client = OrcaAsyncClient._resolve_client()
        async_client.api_key = api_key

    @staticmethod
    def get_api_url() -> str:
        """
        Get the base URL of the Orca API that is currently being used
        """
        client = OrcaClient._resolve_client()
        async_client = OrcaAsyncClient._resolve_client()
        if client.base_url != async_client.base_url:
            raise RuntimeError("The base URL of the sync and async clients do not match")
        return str(client.base_url)

    @staticmethod
    def set_api_url(url: str, check_validity: bool = True):
        """
        Set the base URL for the Orca API

        Note:
            The base URL can also be provided by setting the `ORCA_API_URL` environment variable

        Args:
            url: The base URL to set
            check_validity: Whether to check if there is an API running at the given base URL

        Raises:
            ValueError: if there is no healthy API running at the given base URL and `check_validity` is True
        """
        # check if the base url is reachable before setting it
        if check_validity:
            try:
                httpx.get(url, timeout=1)
            except ConnectError as e:
                raise ValueError(f"No API found at {url}") from e

        sync_client = OrcaClient._resolve_client()
        sync_client.base_url = url

        async_client = OrcaAsyncClient._resolve_client()
        async_client.base_url = url

        # check if the api passes the health check
        if check_validity:
            OrcaCredentials.is_healthy()

    @staticmethod
    def set_api_headers(headers: dict[str, str]):
        """
        Add or override default HTTP headers for all Orca API requests.

        Params:
            headers: Mapping of header names to their string values

        Notes:
            New keys are merged into the existing headers, this will overwrite headers with the
            same name, but leave other headers untouched.
        """
        sync_client = OrcaClient._resolve_client()
        sync_client.headers.update(Headers(headers))

        async_client = OrcaAsyncClient._resolve_client()
        async_client.headers.update(Headers(headers))
